import unittest
import os
from datetime import datetime, timedelta, date
from dateutil.relativedelta import relativedelta
import calendar
import pandas as pd
from src.nemosis import custom_tables, defaults, filters, data_fetch_methods
from pandas._testing import assert_frame_equal
from parameterized import parameterized

recent_test_month = datetime.now().replace(day=1, hour=0, minute=0, second=0, microsecond=0) - relativedelta(months=2)

previous_month = datetime.now().replace(day=1, hour=0, minute=0, second=0, microsecond=0) - relativedelta(months=1)

# Discontinued dates are not precise but should work with current testing dates.
discontinued_after = {
    "TRADINGLOAD": "2022/01/01 00:00:00",
    "TRADINGREGIONSUM": "2022/01/01 00:00:00",
}

test_iterations = [
    [recent_test_month],
    [datetime(year=2018, month=5, day=1)]
]

def get_previous_month(month):
    return month - relativedelta(months=1)

def get_next_month(month):
    return month + relativedelta(months=1)

def last_day_of_month(year, month):
    day_number = date(year, month, calendar.monthrange(year, month)[1]).day
    return str(day_number).zfill(2)

class TestDynamicDataCompilerWithSettlementDateFiltering(unittest.TestCase):
    def setUp(self):
        self.table_names = [
            "BIDDAYOFFER_D",
            "BIDPEROFFER_D",
            "DISPATCHLOAD",
            "DISPATCHCONSTRAINT",
            "DISPATCH_UNIT_SCADA",
            "DISPATCHPRICE",
            "DISPATCHINTERCONNECTORRES",
            "DISPATCHREGIONSUM",
            "TRADINGLOAD",
            "TRADINGPRICE",
            "TRADINGREGIONSUM",
            "TRADINGINTERCONNECT",
            "ROOFTOP_PV_ACTUAL",
        ]

        self.table_types = {
            "DISPATCHLOAD": "DUID",
            "DISPATCHCONSTRAINT": "CONSTRAINTID",
            "DISPATCH_UNIT_SCADA": "DUIDONLY",
            "DISPATCHPRICE": "REGIONID",
            "DISPATCHINTERCONNECTORRES": "INTERCONNECTORID",
            "DISPATCHREGIONSUM": "REGIONID",
            "BIDPEROFFER_D": "DUID-BIDTYPE",
            "BIDDAYOFFER_D": "DUID-BIDTYPE",
            "TRADINGLOAD": "DUIDONLY",
            "TRADINGPRICE": "REGIONIDONLY",
            "TRADINGREGIONSUM": "REGIONIDONLY",
            "TRADINGINTERCONNECT": "INTERCONNECTORIDONLY",
            "ROOFTOP_PV_ACTUAL": "REGIONID-PVVALUETYPE",
        }

        self.table_filters = {
            "DISPATCHLOAD": ["DUID", "INTERVENTION"],
            "DISPATCHCONSTRAINT": ["CONSTRAINTID", "INTERVENTION"],
            "DISPATCH_UNIT_SCADA": ["DUID"],
            "DISPATCHPRICE": ["REGIONID", "INTERVENTION"],
            "DISPATCHINTERCONNECTORRES": ["INTERCONNECTORID", "INTERVENTION"],
            "DISPATCHREGIONSUM": ["REGIONID", "INTERVENTION"],
            "BIDPEROFFER_D": ["DUID", "BIDTYPE"],
            "BIDDAYOFFER_D": ["DUID", "BIDTYPE"],
            "TRADINGLOAD": ["DUID"],
            "TRADINGPRICE": ["REGIONID"],
            "TRADINGREGIONSUM": ["REGIONID"],
            "TRADINGINTERCONNECT": ["INTERCONNECTORID"],
            "ROOFTOP_PV_ACTUAL": ["REGIONID", "TYPE"],
        }

        self.filter_values = {
            "DUID": (["AGLHAL"], [0]),
            "DUIDONLY": (["AGLHAL"],),
            "REGIONID": (["SA1"], [0]),
            "REGIONIDONLY": (["SA1"],),
            "INTERCONNECTORID": (["VIC1-NSW1"], [0]),
            "INTERCONNECTORIDONLY": (["VIC1-NSW1"],),
            "CONSTRAINTID": (["DATASNAP_DFS_Q_CLST"], [0]),
            "DUID-BIDTYPE": (["AGLHAL"], ["ENERGY"]),
            "REGIONID-TYPE": (['NSW1'], ['DAILY']),
            "REGIONID-PVVALUETYPE": (['NSW1'], ['SATELLITE', 'DAILY']),
        }
        self.half_hourly_tables = [
            "TRADINGLOAD",
            "TRADINGPRICE",
            "TRADINGREGIONSUM",
            "TRADINGINTERCONNECT",
            "ROOFTOP_PV_ACTUAL"
        ]
        self.change_to_five_min_table = [
            "TRADINGPRICE",
            "TRADINGINTERCONNECT",
        ]
        self.change_to_five_min_year = 2022

    def _pv_extra_clean_up(self, data):
        pv_types = data['TYPE'].unique()
        if 'DAILY' in pv_types and 'SATELLITE' in pv_types:
            data = data[data['TYPE'] == 'SATELLITE'].copy()
        return data

    @parameterized.expand(test_iterations)
    def test_dispatch_tables_start_of_month(self, test_month):
        start_time = f"{test_month.year}/{test_month.month}/01 00:00:00"
        end_time = f"{test_month.year}/{test_month.month}/01 05:15:00"
        for table in self.table_names:
            if table in discontinued_after and discontinued_after[table] < start_time:
                continue
            print(f"Testing {table} returning values at start of month.")
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            filter_cols = self.table_filters[table]
            cols = [dat_col, *filter_cols]
            expected_length = 63
            expected_number_of_columns = 2
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            if (table in self.half_hourly_tables and
                    not (table in self.change_to_five_min_table and test_month.year > self.change_to_five_min_year)):
                expected_length = 10
                expected_first_time = f"{test_month.year}/{test_month.month}/01 00:30:00"
                expected_first_time = pd.to_datetime(
                    expected_first_time, format="%Y/%m/%d %H:%M:%S"
                )
                expected_last_time = f"{test_month.year}/{test_month.month}/01 05:00:00"
                expected_last_time = pd.to_datetime(
                    expected_last_time, format="%Y/%m/%d %H:%M:%S"
                )
            if "ONLY" not in table_type:
                expected_number_of_columns = 3
            if table == "BIDDAYOFFER_D":
                expected_length = 2
                expected_last_time = f"{test_month.year}/{test_month.month}/01 00:00:00"
                expected_last_time = pd.to_datetime(
                    expected_last_time, format="%Y/%m/%d %H:%M:%S"
                )
                previous_month = get_previous_month(test_month)
                day = last_day_of_month(previous_month.year, previous_month.month)
                expected_first_time = f"{previous_month.year}/{previous_month.month}/{day} 00:00:00"
                expected_first_time = pd.to_datetime(
                    expected_first_time, format="%Y/%m/%d %H:%M:%S"
                )
            print(start_time)
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                select_columns=cols,
                fformat="feather",
                keep_csv=False,
                filter_cols=filter_cols,
                filter_values=self.filter_values[table_type],
            )
            if table == 'ROOFTOP_PV_ACTUAL':
                data = self._pv_extra_clean_up(data)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")

    @parameterized.expand(test_iterations)
    def test_dispatch_tables_end_of_month(self, test_month):
        day = last_day_of_month(test_month.year, test_month.month)
        start_time = f"{test_month.year}/{test_month.month}/{day} 21:00:00"
        following_month = get_next_month(test_month)
        end_time = f"{following_month.year}/{following_month.month}/01 00:00:00"
        for table in self.table_names:
            if table in discontinued_after and discontinued_after[table] < start_time:
                continue
            print("Testing {} returing values at end of month.".format(table))
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            filter_cols = self.table_filters[table]
            cols = [dat_col, *filter_cols]
            expected_length = 36
            expected_number_of_columns = 2
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            if (table in self.half_hourly_tables and
                    not (table in self.change_to_five_min_table and test_month.year > self.change_to_five_min_year)):
                expected_length = 6
                expected_first_time = f"{test_month.year}/{test_month.month}/{day} 21:30:00"
                expected_first_time = pd.to_datetime(
                    expected_first_time, format="%Y/%m/%d %H:%M:%S"
                )
            if "ONLY" not in table_type:
                expected_number_of_columns = 3
            if table == "BIDDAYOFFER_D":
                expected_length = 1
                expected_last_time = expected_first_time.replace(hour=0, minute=0)
                expected_first_time = expected_first_time.replace(hour=0, minute=0)
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                select_columns=cols,
                fformat="feather",
                keep_csv=False,
                filter_cols=filter_cols,
                filter_values=self.filter_values[table_type],
            )
            if table == 'ROOFTOP_PV_ACTUAL':
                data = self._pv_extra_clean_up(data)
            data = data.sort_values(dat_col)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")

    @parameterized.expand(test_iterations)
    def test_dispatch_tables_straddle_2_months(self, test_month):
        day = last_day_of_month(test_month.year, test_month.month)
        start_time = f"{test_month.year}/{test_month.month}/{day} 21:00:00"
        following_month = get_next_month(test_month)
        end_time = f"{following_month.year}/{following_month.month}/01 21:00:00"
        for table in self.table_names:
            if table in discontinued_after and discontinued_after[table] < start_time:
                continue
            print(f"Testing {table} returing values from adjacent months.")
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            filter_cols = self.table_filters[table]
            cols = [dat_col, *filter_cols]
            expected_length = 288
            expected_number_of_columns = 2
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            if (table in self.half_hourly_tables and
                    not (table in self.change_to_five_min_table and test_month.year > self.change_to_five_min_year)):
                expected_length = 48
                expected_first_time = f"{test_month.year}/{test_month.month}/{day} 21:30:00"
                expected_first_time = pd.to_datetime(
                    expected_first_time, format="%Y/%m/%d %H:%M:%S"
                )
            if "ONLY" not in table_type:
                expected_number_of_columns = 3
            if table == "BIDDAYOFFER_D":
                expected_length = 2
                expected_last_time = expected_last_time.replace(hour=0, minute=0)
                expected_first_time = expected_first_time.replace(hour=0, minute=0)
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                select_columns=cols,
                fformat="feather",
                keep_csv=False,
                filter_cols=filter_cols,
                filter_values=self.filter_values[table_type],
            )
            if table == 'ROOFTOP_PV_ACTUAL':
                data = self._pv_extra_clean_up(data)
            data = data.sort_values(dat_col)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")

    @parameterized.expand(test_iterations)
    def test_dispatch_tables_start_of_year(self, test_month):
        start_time = f"{test_month.year}/01/01 00:00:00"
        following_month = get_next_month(test_month)
        end_time = f"{following_month.year}/01/01 01:00:00"
        for table in self.table_names:
            if table in discontinued_after and discontinued_after[table] < start_time:
                continue
            print("Testing {} returing values at start of year.".format(table))
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            filter_cols = self.table_filters[table]
            cols = [dat_col, *filter_cols]
            expected_length = 12
            expected_number_of_columns = 2
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            if (table in self.half_hourly_tables and
                    not (table in self.change_to_five_min_table and test_month.year > self.change_to_five_min_year)):
                expected_length = 2
                expected_first_time = f"{test_month.year}/01/01 00:30:00"
                expected_first_time = pd.to_datetime(
                    expected_first_time, format="%Y/%m/%d %H:%M:%S"
                )
            if "ONLY" not in table_type:
                expected_number_of_columns = 3
            if table == "BIDDAYOFFER_D":
                expected_length = 1
                expected_last_time = expected_last_time.replace(
                    hour=0, minute=0
                ) - timedelta(days=1)
                expected_first_time = expected_first_time.replace(
                    hour=0, minute=0
                ) - timedelta(days=1)
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                select_columns=cols,
                fformat="feather",
                keep_csv=False,
                filter_cols=filter_cols,
                filter_values=self.filter_values[table_type],
            )
            if table == 'ROOFTOP_PV_ACTUAL':
                data = self._pv_extra_clean_up(data)
            data = data.sort_values(dat_col)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")

    @parameterized.expand(test_iterations)
    def test_dispatch_tables_end_of_year(self, test_month):
        start_time = f"{test_month.year - 1}/12/31 23:00:00"
        following_month = get_next_month(test_month)
        end_time = f"{following_month.year}/01/01 00:00:00"
        for table in self.table_names:
            if table in discontinued_after and discontinued_after[table] < start_time:
                continue
            print("Testing {} returing values at end of year.".format(table))
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            filter_cols = self.table_filters[table]
            cols = [dat_col, *filter_cols]
            expected_length = 12
            expected_number_of_columns = 2
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            if (table in self.half_hourly_tables and
                    not (table in self.change_to_five_min_table and test_month.year > self.change_to_five_min_year)):
                expected_length = 2
                expected_first_time = f"{test_month.year - 1}/12/31 23:30:00"
                expected_first_time = pd.to_datetime(
                    expected_first_time, format="%Y/%m/%d %H:%M:%S"
                )
            if "ONLY" not in table_type:
                expected_number_of_columns = 3
            if table == "BIDDAYOFFER_D":
                expected_length = 1
                expected_first_time = expected_first_time.replace(hour=0, minute=0)
                expected_last_time = expected_first_time
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                select_columns=cols,
                fformat="feather",
                keep_csv=False,
                filter_cols=filter_cols,
                filter_values=self.filter_values[table_type],
            )
            if table == 'ROOFTOP_PV_ACTUAL':
                data = self._pv_extra_clean_up(data)
            data = data.sort_values(dat_col)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")

    @parameterized.expand(test_iterations)
    def test_dispatch_tables_straddle_years(self, test_month):
        start_time = f"{test_month.year - 1}/12/31 23:00:00"
        following_month = get_next_month(test_month)
        end_time = f"{following_month.year}/01/01 01:00:00"
        for table in self.table_names:
            if table in discontinued_after and discontinued_after[table] < start_time:
                continue
            print(f"Testing {table} returning values from adjacent years.")
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            filter_cols = self.table_filters[table]
            cols = [dat_col, *filter_cols]
            expected_length = 24
            expected_number_of_columns = 2
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            if (table in self.half_hourly_tables and
                    not (table in self.change_to_five_min_table and test_month.year > self.change_to_five_min_year)):
                expected_length = 4
                expected_first_time = f"{test_month.year - 1}/12/31 23:30:00"
                expected_first_time = pd.to_datetime(
                    expected_first_time, format="%Y/%m/%d %H:%M:%S"
                )
            if "ONLY" not in table_type:
                expected_number_of_columns = 3
            if table == "BIDDAYOFFER_D":
                expected_length = 1
                expected_first_time = expected_first_time.replace(hour=0, minute=0)
                expected_last_time = expected_first_time
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                select_columns=cols,
                fformat="feather",
                keep_csv=False,
                filter_cols=filter_cols,
                filter_values=self.filter_values[table_type],
            )
            if table == 'ROOFTOP_PV_ACTUAL':
                data = self._pv_extra_clean_up(data)
            data = data.sort_values(dat_col)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")


class TestDynamicDataCompilerWithSettlementDateFiltering2021OfferData(
    unittest.TestCase
):
    """
    Testing during 2021 because this is when AEMO stopped storing BIDDAYOFFER_D and BIDPEROFFER_D with the other mms
    monthly files. Testing at this time checks that retreiving from the BIDMOVECOMPLETE files works.
    """
    def setUp(self):
        self.table_names = ["BIDDAYOFFER_D", "BIDPEROFFER_D"]

        self.table_types = {
            "BIDPEROFFER_D": "DUID-BIDTYPE",
            "BIDDAYOFFER_D": "DUID-BIDTYPE",
        }

        self.table_filters = {
            "BIDPEROFFER_D": ["DUID", "BIDTYPE"],
            "BIDDAYOFFER_D": ["DUID", "BIDTYPE"],
        }

        # Filter for bids at the start of the 2021-06-01 file and the end of the 2021-05-31, to make sure that we aren't
        # skipping any of the data file rows.
        self.filter_values = {
            "DUID-BIDTYPE": (
                ["ADPBA1G", "ARWF1", "YWPS4", "YWPS4"],
                ["ENERGY", "RAISEREG", "RAISE60SEC"],
            )
        }

    @staticmethod
    def restrictive_filter(data):
        data = data[
            (data["DUID"] == "ADPBA1G") & (data["BIDTYPE"] == "ENERGY")
            | (data["DUID"] == "ARWF1") & (data["BIDTYPE"] == "ENERGY")
            | (data["DUID"] == "YWPS4") & (data["BIDTYPE"] == "RAISEREG")
            | (data["DUID"] == "YWPS4") & (data["BIDTYPE"] == "RAISE60SEC")
        ]
        return data

    def test_dispatch_tables_start_of_month(self):
        start_time = "2021/09/01 00:00:00"
        end_time = "2021/09/01 05:15:00"
        for table in self.table_names:
            print(f"Testing {table} returning values at start of month one.")
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            filter_cols = self.table_filters[table]
            cols = [dat_col, *filter_cols]
            expected_length = 63 * 4
            expected_number_of_columns = 3
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            if table == "BIDDAYOFFER_D":
                expected_length = 2 * 4
                expected_last_time = "2021/09/01 00:00:00"
                expected_last_time = pd.to_datetime(
                    expected_last_time, format="%Y/%m/%d %H:%M:%S"
                )
                expected_first_time = "2021/08/31 00:00:00"
                expected_first_time = pd.to_datetime(
                    expected_first_time, format="%Y/%m/%d %H:%M:%S"
                )
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                select_columns=cols,
                fformat="feather",
                keep_csv=False,
                filter_cols=filter_cols,
                filter_values=self.filter_values[table_type],
            )
            data = self.restrictive_filter(data)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")

    def test_dispatch_tables_start_of_month_previous_market_day_but_not_start_calendar_month(
        self,
    ):
        start_time = "2021/09/05 03:00:00"
        end_time = "2021/09/05 03:15:00"
        for table in self.table_names:
            print(f"Testing {table} returning values at start of month two.")
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            filter_cols = self.table_filters[table]
            cols = [dat_col, *filter_cols]
            expected_length = 3 * 4
            expected_number_of_columns = 3
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            if table == "BIDDAYOFFER_D":
                expected_length = 1 * 4
                expected_last_time = "2021/09/04 00:00:00"
                expected_last_time = pd.to_datetime(
                    expected_last_time, format="%Y/%m/%d %H:%M:%S"
                )
                expected_first_time = "2021/09/04 00:00:00"
                expected_first_time = pd.to_datetime(
                    expected_first_time, format="%Y/%m/%d %H:%M:%S"
                )
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                select_columns=cols,
                fformat="feather",
                keep_csv=False,
                filter_cols=filter_cols,
                filter_values=self.filter_values[table_type],
            )
            data = self.restrictive_filter(data)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")

    def test_dispatch_tables_start_of_month_previous_market_day_and_first_market_day_but_not_start_calendar_month(
        self,
    ):
        start_time = "2021/09/01 03:00:00"
        end_time = "2021/09/01 05:00:00"
        for table in self.table_names:
            print(f"Testing {table} returning values at start of month two.")
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            filter_cols = self.table_filters[table]
            cols = [dat_col, *filter_cols]
            expected_length = 24 * 4
            expected_number_of_columns = 3
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            if table == "BIDDAYOFFER_D":
                expected_length = 2 * 4
                expected_last_time = "2021/09/01 00:00:00"
                expected_last_time = pd.to_datetime(
                    expected_last_time, format="%Y/%m/%d %H:%M:%S"
                )
                expected_first_time = "2021/08/31 00:00:00"
                expected_first_time = pd.to_datetime(
                    expected_first_time, format="%Y/%m/%d %H:%M:%S"
                )
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                select_columns=cols,
                fformat="feather",
                keep_csv=False,
                filter_cols=filter_cols,
                filter_values=self.filter_values[table_type],
            )
            data = self.restrictive_filter(data)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")

    def test_dispatch_tables_start_of_month_first_market_day_but_not_start_calendar_month(
        self,
    ):
        start_time = "2021/09/01 04:00:00"
        end_time = "2021/09/01 05:00:00"
        for table in self.table_names:
            print(f"Testing {table} returning values at start of month two.")
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            filter_cols = self.table_filters[table]
            cols = [dat_col, *filter_cols]
            expected_length = 12 * 4
            expected_number_of_columns = 3
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            if table == "BIDDAYOFFER_D":
                expected_length = 1 * 4
                expected_last_time = "2021/09/01 00:00:00"
                expected_last_time = pd.to_datetime(
                    expected_last_time, format="%Y/%m/%d %H:%M:%S"
                )
                expected_first_time = "2021/09/01 00:00:00"
                expected_first_time = pd.to_datetime(
                    expected_first_time, format="%Y/%m/%d %H:%M:%S"
                )
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                select_columns=cols,
                fformat="feather",
                keep_csv=False,
                filter_cols=filter_cols,
                filter_values=self.filter_values[table_type],
            )
            data = self.restrictive_filter(data)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")

    def test_dispatch_tables_end_of_month(self):
        start_time = "2021/09/30 21:00:00"
        end_time = "2021/10/01 00:00:00"
        for table in self.table_names:
            print("Testing {} returing values at end of month.".format(table))
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            filter_cols = self.table_filters[table]
            cols = [dat_col, *filter_cols]
            expected_length = 36 * 4
            expected_number_of_columns = 3
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            if table == "BIDDAYOFFER_D":
                expected_length = 1 * 4
                expected_last_time = expected_first_time.replace(hour=0, minute=0)
                expected_first_time = expected_first_time.replace(hour=0, minute=0)
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                select_columns=cols,
                fformat="feather",
                keep_csv=False,
                filter_cols=filter_cols,
                filter_values=self.filter_values[table_type],
            )
            data = self.restrictive_filter(data)
            data = data.sort_values(dat_col)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")

    def test_dispatch_tables_straddle_2_months(self):
        start_time = "2021/09/30 21:00:00"
        end_time = "2021/10/01 21:00:00"
        for table in self.table_names:
            print(f"Testing {table} returing values from adjacent months.")
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            filter_cols = self.table_filters[table]
            cols = [dat_col, *filter_cols]
            expected_length = 240 * 4  # This should be 288 but there data missing in file AEMO published.
            expected_number_of_columns = 3
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            if table == "BIDDAYOFFER_D":
                expected_length = 2 * 4
                expected_last_time = expected_last_time.replace(hour=0, minute=0)
                expected_first_time = expected_first_time.replace(hour=0, minute=0)
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                select_columns=cols,
                fformat="feather",
                keep_csv=False,
                filter_cols=filter_cols,
                filter_values=self.filter_values[table_type],
            )
            data = self.restrictive_filter(data)
            data = data.sort_values(dat_col)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")

    def test_dispatch_tables_start_of_year(self):
        start_time = "2022/01/01 00:00:00"
        end_time = "2022/01/01 01:00:00"
        for table in self.table_names:
            print("Testing {} returing values at start of year.".format(table))
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            filter_cols = self.table_filters[table]
            cols = [dat_col, *filter_cols]
            expected_length = 12 * 4
            expected_number_of_columns = 3
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            if table == "BIDDAYOFFER_D":
                expected_length = 1 * 4
                expected_last_time = expected_last_time.replace(
                    hour=0, minute=0
                ) - timedelta(days=1)
                expected_first_time = expected_first_time.replace(
                    hour=0, minute=0
                ) - timedelta(days=1)
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                select_columns=cols,
                fformat="feather",
                keep_csv=False,
                filter_cols=filter_cols,
                filter_values=self.filter_values[table_type],
            )
            data = self.restrictive_filter(data)
            data = data.sort_values(dat_col)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")

    def test_dispatch_tables_end_of_year(self):
        start_time = "2021/12/31 23:00:00"
        end_time = "2022/01/01 00:00:00"
        for table in self.table_names:
            print("Testing {} returing values at end of year.".format(table))
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            filter_cols = self.table_filters[table]
            cols = [dat_col, *filter_cols]
            expected_length = 12 * 4
            expected_number_of_columns = 3
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            if table == "BIDDAYOFFER_D":
                expected_length = 1 * 4
                expected_first_time = expected_first_time.replace(hour=0, minute=0)
                expected_last_time = expected_first_time
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                select_columns=cols,
                fformat="feather",
                keep_csv=False,
                filter_cols=filter_cols,
                filter_values=self.filter_values[table_type],
            )
            data = self.restrictive_filter(data)
            data = data.sort_values(dat_col)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")

    def test_dispatch_tables_straddle_years(self):
        start_time = "2021/12/31 23:00:00"
        end_time = "2022/01/01 01:00:00"
        for table in self.table_names:
            print(f"Testing {table} returning values from adjacent years.")
            dat_col = defaults.primary_date_columns[table]
            table_type = self.table_types[table]
            filter_cols = self.table_filters[table]
            cols = [dat_col, *filter_cols]
            expected_length = 24 * 4
            expected_number_of_columns = 3
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            if table == "BIDDAYOFFER_D":
                expected_length = 1 * 4
                expected_first_time = expected_first_time.replace(hour=0, minute=0)
                expected_last_time = expected_first_time
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                select_columns=cols,
                fformat="feather",
                keep_csv=False,
                filter_cols=filter_cols,
                filter_values=self.filter_values[table_type],
            )
            data = self.restrictive_filter(data)
            data = data.sort_values(dat_col)
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")


class TestDynamicDataCompilerWithSettlementDateFilteringNextDayTables(
    unittest.TestCase
):

    def setUp(self):
        self.table_names = [
            "DAILY_REGION_SUMMARY",
            "NEXT_DAY_DISPATCHLOAD",
            "INTERMITTENT_GEN_SCADA"
        ]

        self.table_filters = {
            "DAILY_REGION_SUMMARY": ["REGIONID"],
            "NEXT_DAY_DISPATCHLOAD": ["DUID"],
            "INTERMITTENT_GEN_SCADA": ["DUID", "SCADA_TYPE"]
        }

        # Filter for bids at the start of the 2021-06-01 file and the end of the 2021-05-31, to make sure that we aren't
        # skipping any of the data file rows.
        self.filter_values = {
            "DAILY_REGION_SUMMARY": (
                ["NSW1"],
            ),
            "NEXT_DAY_DISPATCHLOAD": (
                ['AGLHAL'],
            ),
            "INTERMITTENT_GEN_SCADA": (
                ['ADPPV1'],
                ['ELAV']
            )
        }

    def test_dispatch_tables_start_of_month(self):
        test_month = previous_month
        start_time = f"{test_month.year}/{test_month.month}/01 00:00:00"
        end_time = f"{test_month.year}/{test_month.month}/01 05:15:00"
        for table in self.table_names:
            print(f"Testing {table} returning values at start of month one.")
            dat_col = defaults.primary_date_columns[table]
            expected_length = 63 * 1
            expected_number_of_columns = len(defaults.table_columns[table])
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                fformat="feather",
                keep_csv=True,
                filter_cols=self.table_filters[table],
                filter_values=self.filter_values[table]
            )
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")

    def test_dispatch_tables_middle_of_month_and_day(self):
        test_month = previous_month
        start_time = f"{test_month.year}/{test_month.month}/05 12:00:00"
        end_time = f"{test_month.year}/{test_month.month}/05 17:15:00"
        for table in self.table_names:
            print(f"Testing {table} returning values at start of month one.")
            dat_col = defaults.primary_date_columns[table]
            expected_length = 63 * 1
            expected_number_of_columns = len(defaults.table_columns[table])
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                fformat="feather",
                keep_csv=True,
                filter_cols=self.table_filters[table],
                filter_values=self.filter_values[table]
            )
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")

    def test_dispatch_tables_start_market_day(self):
        test_month = previous_month
        start_time = f"{test_month.year}/{test_month.month}/05 04:00:00"
        end_time = f"{test_month.year}/{test_month.month}/05 04:05:00"
        for table in self.table_names:
            print(f"Testing {table} returning values at start of month one.")
            dat_col = defaults.primary_date_columns[table]
            expected_length = 1
            expected_number_of_columns = len(defaults.table_columns[table])
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                fformat="feather",
                keep_csv=True,
                filter_cols=self.table_filters[table],
                filter_values=self.filter_values[table]
            )
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")

    def test_dispatch_tables_end_market_day(self):
        test_month = previous_month
        start_time = f"{test_month.year}/{test_month.month}/05 03:55:00"
        end_time = f"{test_month.year}/{test_month.month}/05 04:00:00"
        for table in self.table_names:
            print(f"Testing {table} returning values at start of month one.")
            dat_col = defaults.primary_date_columns[table]
            expected_length = 1
            expected_number_of_columns = len(defaults.table_columns[table])
            expected_first_time = pd.to_datetime(
                start_time, format="%Y/%m/%d %H:%M:%S"
            ) + timedelta(minutes=5)
            expected_last_time = pd.to_datetime(end_time, format="%Y/%m/%d %H:%M:%S")
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                fformat="feather",
                keep_csv=True,
                filter_cols=self.table_filters[table],
                filter_values=self.filter_values[table]
            )
            data = data.reset_index(drop=True)
            self.assertEqual(expected_length, data.shape[0])
            self.assertEqual(expected_number_of_columns, data.shape[1])
            self.assertEqual(expected_first_time, data[dat_col][0])
            self.assertEqual(expected_last_time, data[dat_col].iloc[-1])
            print("Passed")


class TestDynamicDataCompilerWithEffectiveDateFiltering(unittest.TestCase):
    def setUp(self):
        self.table_names = [
            "GENCONDATA",
            "SPDREGIONCONSTRAINT",
            "SPDCONNECTIONPOINTCONSTRAINT",
            "SPDINTERCONNECTORCONSTRAINT",
        ]

    @parameterized.expand([
        [recent_test_month],
        [datetime(year=2018, month=5, day=1)]
    ])
    def test_filtering_for_one_interval_returns(self, test_month):
        start_time = f"{test_month.year}/{test_month.month}/20 23:00:00"
        end_time = f"{test_month.year}/{test_month.month}/20 23:05:00"
        for table in self.table_names:
            print("Testing {} returing values for 1 interval.".format(table))
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                fformat="feather",
                keep_csv=False,
                select_columns=defaults.table_primary_keys[table],
            )
            group_cols = [
                col
                for col in defaults.table_primary_keys[table]
                if col != "EFFECTIVEDATE"
            ]
            contains_duplicates = data.duplicated(group_cols).any()
            self.assertEqual(False, contains_duplicates)
            not_empty = data.shape[0] > 0
            self.assertEqual(True, not_empty)
            print("Passed")


class TestCacheCompiler(unittest.TestCase):
    def setUp(self):
        self.table_names = [
            "BIDDAYOFFER_D",
            "BIDPEROFFER_D",
            "DISPATCHLOAD",
            "DISPATCHCONSTRAINT",
            "DISPATCH_UNIT_SCADA",
            "DISPATCHPRICE",
            "DISPATCHINTERCONNECTORRES",
            "DISPATCHREGIONSUM",
            "TRADINGLOAD",
            "TRADINGPRICE",
            "TRADINGREGIONSUM",
            "TRADINGINTERCONNECT",
        ]
        self.id_cols = {
            "DISPATCHLOAD": "DUID",
            "DISPATCHCONSTRAINT": "CONSTRAINTID",
            "DISPATCH_UNIT_SCADA": "DUID",
            "DISPATCHPRICE": "REGIONID",
            "DISPATCHINTERCONNECTORRES": "INTERCONNECTORID",
            "DISPATCHREGIONSUM": "REGIONID",
            "BIDPEROFFER_D": "DUID",
            "BIDDAYOFFER_D": "DUID",
            "TRADINGLOAD": "DUID",
            "TRADINGPRICE": "REGIONID",
            "TRADINGREGIONSUM": "REGIONID",
            "TRADINGINTERCONNECT": "INTERCONNECTORID",
        }

    def test_caching_and_typing_works_feather(self):
        start_time = "2018/02/20 23:00:00"
        end_time = "2018/02/20 23:30:00"
        for table in self.table_names:
            dat_col = defaults.primary_date_columns[table]
            id_col = self.id_cols[table]
            print(f"Testing {table} returing values for 1 interval.")
            data_fetch_methods.cache_compiler(
                start_time, end_time, table, defaults.raw_data_cache, fformat="feather"
            )
            data = data_fetch_methods.dynamic_data_compiler(
                start_time, end_time, table, defaults.raw_data_cache, fformat="feather"
            )
            dat_col_type = data[dat_col].dtype
            id_col_type = data[id_col].dtype
            not_empty = data.shape[0] > 0
            not_typed = all(data.dtypes == "object")
            self.assertTrue(not_empty)
            self.assertFalse(not_typed)
            self.assertEqual(dat_col_type, "<M8[ns]")
            self.assertEqual(id_col_type, "object")
            print("Passed")

    def test_caching_and_typing_works_parquet(self):
        start_time = "2018/02/20 23:00:00"
        end_time = "2018/02/20 23:30:00"
        for table in self.table_names:
            dat_col = defaults.primary_date_columns[table]
            id_col = self.id_cols[table]
            print("Testing {} returing values for 1 interval.".format(table))
            data_fetch_methods.cache_compiler(
                start_time, end_time, table, defaults.raw_data_cache, fformat="parquet"
            )
            data = data_fetch_methods.dynamic_data_compiler(
                start_time, end_time, table, defaults.raw_data_cache, fformat="parquet"
            )
            dat_col_type = data[dat_col].dtype
            id_col_type = data[id_col].dtype
            not_empty = data.shape[0] > 0
            not_typed = all(data.dtypes == "object")
            self.assertTrue(not_empty)
            self.assertFalse(not_typed)
            self.assertEqual(dat_col_type, "<M8[ns]")
            self.assertEqual(id_col_type, "object")
            print("Passed")

    def test_caching_with_select_columns_works(self):
        start_time = "2018/02/20 23:00:00"
        end_time = "2018/02/20 23:30:00"
        table = "DISPATCHPRICE"
        print("Testing {} returing values for 1 interval.".format(table))
        data_fetch_methods.cache_compiler(
            start_time,
            end_time,
            table,
            defaults.raw_data_cache,
            fformat="parquet",
            rebuild=True,
        )
        data_fetch_methods.cache_compiler(
            start_time,
            end_time,
            table,
            defaults.raw_data_cache,
            fformat="parquet",
            select_columns=["SETTLEMENTDATE", "REGIONID"],
            rebuild=True,
        )
        data = pd.read_parquet(
            os.path.join(
                defaults.raw_data_cache, "PUBLIC_DVD_DISPATCHPRICE_201802010000.parquet"
            )
        )
        data_fetch_methods.cache_compiler(
            start_time,
            end_time,
            table,
            defaults.raw_data_cache,
            fformat="parquet",
            rebuild=True,
        )
        got_columns = list(data.columns)
        expected_columns = ["SETTLEMENTDATE", "REGIONID"]
        self.assertSequenceEqual(got_columns, expected_columns)
        print("Passed")


class TestDynamicDataCompilerWithStartDateFiltering(unittest.TestCase):
    def setUp(self):
        self.table_names = ["DUDETAILSUMMARY"]

    @parameterized.expand([
        [recent_test_month],
        [datetime(year=2018, month=5, day=1)]
    ])
    def test_filtering_for_one_interval_returns(self, test_month):
        start_time = f"{test_month.year}/{test_month.month}/20 23:00:00"
        end_time = f"{test_month.year}/{test_month.month}/20 23:05:00"
        for table in self.table_names:
            print("Testing {} returing values for 1 interval.".format(table))
            data = data_fetch_methods.dynamic_data_compiler(
                start_time,
                end_time,
                table,
                defaults.raw_data_cache,
                fformat="feather",
                keep_csv=False,
                select_columns=defaults.table_primary_keys[table],
            )
            group_cols = [
                col for col in defaults.table_primary_keys[table] if col != "START_DATE"
            ]
            contains_duplicates = data.duplicated(group_cols).any()
            self.assertEqual(False, contains_duplicates)
            not_empty = data.shape[0] > 0
            self.assertEqual(True, not_empty)
            print("Passed")


class TestDynamicDataCompilerWithLastChangedFiltering(unittest.TestCase):
    @parameterized.expand([
        [recent_test_month],
        [datetime(year=2018, month=5, day=1)]
    ])
    def test_that_a_narrow_time_range_returns_one_entry_per_participantid(self, test_month):
        start_time = f"{test_month.year}/{test_month.month}/01 00:00:00"
        end_time = f"{test_month.year}/{test_month.month}/01 00:05:00"
        table = "PARTICIPANT"
        cols = ["PARTICIPANTID", "NAME", "LASTCHANGED"]
        data = data_fetch_methods.dynamic_data_compiler(
            start_time,
            end_time,
            table,
            defaults.raw_data_cache,
            select_columns=cols,
        )
        data = data.groupby(['PARTICIPANTID'], as_index=False).count()
        assert (data['NAME'] == 1).all()
        print("Passed")

    def test_that_a_wide_time_range_returns_more_than_one_entry_for_at_least_some_participantids(self):
        table = "PARTICIPANT"
        data = data_fetch_methods.dynamic_data_compiler(
            "2013/01/01 00:00:00",
            "2023/01/01 00:00:00",
            table,
            defaults.raw_data_cache,
        )
        data = data.groupby(['PARTICIPANTID'], as_index=False).count()
        assert (data['NAME'] > 1).any()
        print("Passed")

    def test_filter_of_raw_data(self):
        table = "PARTICIPANT"
        start_time_str = defaults.nem_data_model_start_time
        end_time_str = "2023/01/01 00:00:00"
        start_search = defaults.nem_data_model_start_time
        start_time = datetime.strptime(start_time_str, "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime(end_time_str, "%Y/%m/%d %H:%M:%S")
        start_search = datetime.strptime(start_search, "%Y/%m/%d %H:%M:%S")
        data = data_fetch_methods.dynamic_data_compiler(
            start_time_str,
            end_time_str,
            table,
            defaults.raw_data_cache,
        )
        raw_data = data_fetch_methods._dynamic_data_fetch_loop(
            start_search,
            start_time,
            end_time,
            table,
            defaults.raw_data_cache,
            select_columns=['PARTICIPANTID', 'NAME', 'PARTICIPANTCLASSID', 'LASTCHANGED'],
            date_filter=filters.filter_on_last_changed,
            fformat='feather',
            keep_csv=True,
            rebuild=False)
        raw_data = pd.concat(raw_data)
        raw_data = raw_data.groupby(['PARTICIPANTID', 'LASTCHANGED'], as_index=False).count().\
            groupby(['PARTICIPANTID'], as_index=False).count().sort_values('PARTICIPANTID').reset_index(drop=True)
        data = data.groupby(['PARTICIPANTID', 'LASTCHANGED'], as_index=False).count().\
            groupby(['PARTICIPANTID'], as_index=False).count().sort_values('PARTICIPANTID').reset_index(drop=True)
        assert_frame_equal(raw_data, data)


class TestFCAS4SecondData(unittest.TestCase):
    def setUp(self):
        self.start_day = (datetime.now() - timedelta(30)).replace(
            hour=0, minute=0, second=0, microsecond=0
        )
        self.start_month = self.start_day.replace(
            day=1, hour=0, minute=0, second=0, microsecond=0
        )
        self.start_year = self.start_day.replace(
            month=1, day=1, hour=0, minute=0, second=0, microsecond=0
        )
        pass

    def test_dispatch_tables_start_of_month(self):
        table = "FCAS_4_SECOND"
        start_time = self.start_day
        minute_offset = 5
        end_time = start_time + timedelta(minutes=minute_offset)
        start_str = start_time.strftime("%Y/%m/%d %H:%M:%S")
        end_str = end_time.strftime("%Y/%m/%d %H:%M:%S")

        print("Testing {} returing values at start of month.".format(table))
        dat_col = defaults.primary_date_columns[table]
        cols = [dat_col, "ELEMENTNUMBER", "VARIABLENUMBER"]
        # expected length assumes first data point is at 00:00:00
        expected_length = 15 * minute_offset
        length_check = False
        expected_number_of_columns = 3
        data = data_fetch_methods.dynamic_data_compiler(
            start_str,
            end_str,
            table,
            defaults.raw_data_cache,
            fformat="feather",
            keep_csv=False,
            select_columns=cols,
        )
        length_array = data[dat_col].drop_duplicates()
        if length_array.shape[0] == expected_length:
            length_check = True
        # case if first data point is 00:00:01/02/03
        elif length_array.shape[0] == expected_length - 1:
            length_check = True
        self.assertTrue(length_check)
        self.assertEqual(expected_number_of_columns, data.shape[1])
        print("Passed")

    def test_fcas_tables_end_of_month(self):
        table = "FCAS_4_SECOND"
        minute_offset = 5
        start_time = self.start_month - timedelta(minutes=minute_offset)
        end_time = start_time + timedelta(minutes=minute_offset * 2)
        start_str = start_time.strftime("%Y/%m/%d %H:%M:%S")
        end_str = end_time.strftime("%Y/%m/%d %H:%M:%S")

        print("Testing {} returing values at end of month.".format(table))
        dat_col = defaults.primary_date_columns[table]
        cols = [dat_col, "ELEMENTNUMBER", "VARIABLENUMBER"]
        # expected length assumes first data point is at 00:00:00
        expected_length = 15 * (minute_offset * 2)
        length_check = False
        expected_number_of_columns = 3
        data = data_fetch_methods.dynamic_data_compiler(
            start_str,
            end_str,
            table,
            defaults.raw_data_cache,
            fformat="feather",
            keep_csv=False,
            select_columns=cols,
        )
        length_array = data[dat_col].drop_duplicates()
        if length_array.shape[0] == expected_length:
            length_check = True
        # case if first data point is 00:00:01/02/03
        elif length_array.shape[0] == expected_length - 1:
            length_check = True
        print(length_array.shape[0])
        self.assertTrue(length_check)
        self.assertEqual(expected_number_of_columns, data.shape[1])
        print("Passed")

    @unittest.skipIf(
        (datetime.now() - datetime(year=datetime.now().year, month=1, day=1)).days > 60,
        "start of year data not available: > 60 days ago",
    )
    def test_fcas_tables_end_of_year(self):
        table = "FCAS_4_SECOND"
        minute_offset = 5
        start_time = self.start_year - timedelta(minutes=minute_offset)
        end_time = start_time + timedelta(minutes=minute_offset * 2)
        start_str = start_time.strftime("%Y/%m/%d %H:%M:%S")
        end_str = end_time.strftime("%Y/%m/%d %H:%M:%S")

        print("Testing {} returing values at end of year.".format(table))
        dat_col = defaults.primary_date_columns[table]
        cols = [dat_col, "ELEMENTNUMBER", "VARIABLENUMBER"]
        expected_length = 15 * (minute_offset * 2)
        length_check = False
        expected_number_of_columns = 3
        data = data_fetch_methods.dynamic_data_compiler(
            start_str,
            end_str,
            table,
            defaults.raw_data_cache,
            select_columns=cols,
            fformat="feather",
            keep_csv=False,
        )
        length_array = data[dat_col].drop_duplicates()
        if length_array.shape[0] == expected_length:
            length_check = True
        # case if first data point is 00:00:01/02/03
        elif length_array.shape[0] == expected_length - 1:
            length_check = True
        self.assertTrue(length_check)
        self.assertEqual(expected_number_of_columns, data.shape[1])
        print("Passed")


class TestStaticTables(unittest.TestCase):
    def setUp(self):
        pass

    def test_fcas_elements_table(self):
        start_time = "2017/12/31 23:55:04"
        end_time = "2018/01/01 00:05:00"
        table = "ELEMENTS_FCAS_4_SECOND"
        cols = ["ELEMENTNUMBER", "EMSNAME"]
        filter_cols = ("ELEMENTNUMBER",)
        func = data_fetch_methods._static_table_wrapper_for_gui
        data = func(
            start_time,
            end_time,
            table,
            defaults.raw_data_cache,
            select_columns=cols,
            filter_cols=filter_cols,
            filter_values=(["1"],),
        )
        expected_length = 1
        expected_number_of_columns = 2
        self.assertEqual(expected_length, data.shape[0])
        self.assertEqual(expected_number_of_columns, data.shape[1])
        print("Passed")

    def test_fcas_variable_table(self):
        start_time = "2018/12/31 23:55:04"
        end_time = "2018/01/01 00:05:00"
        table = "VARIABLES_FCAS_4_SECOND"
        cols = ["VARIABLENUMBER", "VARIABLETYPE"]
        filter_cols = ("VARIABLENUMBER",)
        func = data_fetch_methods._static_table_wrapper_for_gui
        data = func(
            start_time,
            end_time,
            table,
            defaults.raw_data_cache,
            select_columns=cols,
            filter_cols=filter_cols,
            filter_values=(["2"],),
        )
        expected_length = 1
        expected_number_of_columns = 2
        self.assertEqual(expected_length, data.shape[0])
        self.assertEqual(expected_number_of_columns, data.shape[1])
        print("Passed")

    def test_registration_list(self):
        table = "Generators and Scheduled Loads"
        cols = ["DUID", "Technology Type - Primary"]
        filter_cols = ("DUID",)
        data = data_fetch_methods.static_table_xl(
            table,
            defaults.raw_data_cache,
            select_columns=cols,
            filter_cols=filter_cols,
            filter_values=(["AGLHAL"],),
        )
        expected_length = 1
        expected_number_of_columns = 2
        self.assertEqual(expected_length, data.shape[0])
        self.assertEqual(expected_number_of_columns, data.shape[1])
        print("Passed")


class TestCustomTables(unittest.TestCase):
    def setUp(self):
        pass

    @unittest.skipIf(
        (datetime.now() - datetime(year=datetime.now().year, month=1, day=1)).days > 60,
        "start of year data not available: > 60 days ago",
    )
    def test_dispatch_tables_straddle_years(self):
        table = "FCAS_4_SECOND"
        minute_offset = 5
        start_time = self.start_year - timedelta(minutes=minute_offset)
        end_time = start_time + timedelta(minutes=minute_offset * 2)

        print("Testing custom table {}.".format(table))
        data = custom_tables.fcas4s_scada_match(
            start_time, end_time, table, defaults.raw_data_cache
        )
        data = data.reset_index(drop=True)
        contains_duplicates = data.duplicated(["MARKETNAME"]).any()
        self.assertEqual(False, contains_duplicates)
        contains_duplicates = data.duplicated(["ELEMENTNUMBER"]).any()
        self.assertEqual(False, contains_duplicates)
        not_empty = data.shape[0] > 0
        self.assertEqual(True, not_empty)
        print("Passed")



import pandas as pd
from datetime import timedelta, datetime
import math
import numpy as np
from nemosis import defaults, data_fetch_methods, filters


def fcas4s_scada_match(
    start_time,
    end_time,
    table_name,
    raw_data_location,
    select_columns=None,
    filter_cols=None,
    filter_values=None,
):

    # Pull in the 4 second fcas data.
    table_name_fcas4s = "FCAS_4_SECOND"
    fcas4s = data_fetch_methods.dynamic_data_compiler(
        start_time, end_time, table_name_fcas4s, raw_data_location
    )
    # Pull in the 4 second fcas variable types.
    table_name_variable_types = "VARIABLES_FCAS_4_SECOND"
    fcas4s_variable_types = data_fetch_methods.static_table(
        table_name_variable_types, raw_data_location
    )

    # Select the variable types that measure MW on an interconnector and Gen_MW from a dispatch unit.
    fcas4s_variable_types = fcas4s_variable_types[
        fcas4s_variable_types["VARIABLETYPE"].isin(["MW", "Gen_MW"])
    ]
    fcas4s = fcas4s[
        fcas4s["VARIABLENUMBER"].isin(fcas4s_variable_types["VARIABLENUMBER"])
    ]

    # Select just the fcas 4 second data variable columns that we need.
    fcas4s = fcas4s.loc[:, ("TIMESTAMP", "ELEMENTNUMBER", "VALUE")]

    # Convert the fcas MW measured values to numeric type.
    fcas4s["VALUE"] = pd.to_numeric(fcas4s["VALUE"])

    # Rename the 4 second measurements to the timestamp of the start of the 5 min interval i.e round down to nearest
    # 5 min interval.
    fcas4s = fcas4s[
        (fcas4s["TIMESTAMP"].dt.minute.isin(list(range(0, 60, 5))))
        & (fcas4s["TIMESTAMP"].dt.second < 20)
    ]
    fcas4s["TIMESTAMP"] = fcas4s["TIMESTAMP"].apply(
        lambda dt: datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute)
    )

    # Pull in the dispatch unit scada data.
    table_name_scada = "DISPATCH_UNIT_SCADA"
    scada = data_fetch_methods.dynamic_data_compiler(
        start_time, end_time, table_name_scada, raw_data_location
    )
    scada["SETTLEMENTDATE"] = scada["SETTLEMENTDATE"] - timedelta(minutes=5)
    scada = scada.loc[:, ("SETTLEMENTDATE", "DUID", "SCADAVALUE")]
    scada.columns = ["SETTLEMENTDATE", "MARKETNAME", "SCADAVALUE"]
    scada["SCADAVALUE"] = pd.to_numeric(scada["SCADAVALUE"])

    # Pull in the interconnector scada data and use the intervention records where the exist.
    table_name_inter_flow = "DISPATCHINTERCONNECTORRES"
    inter_flows = data_fetch_methods.dynamic_data_compiler(
        start_time, end_time, table_name_inter_flow, raw_data_location
    )
    inter_flows["METEREDMWFLOW"] = pd.to_numeric(inter_flows["METEREDMWFLOW"])
    inter_flows = inter_flows.sort_values("INTERVENTION")
    inter_flows = inter_flows.groupby(
        ["SETTLEMENTDATE", "INTERCONNECTORID"], as_index=False
    ).last()
    inter_flows = inter_flows.loc[
        :, ("SETTLEMENTDATE", "INTERCONNECTORID", "METEREDMWFLOW")
    ]
    inter_flows["SETTLEMENTDATE"] = inter_flows["SETTLEMENTDATE"] - timedelta(minutes=5)
    inter_flows.columns = ["SETTLEMENTDATE", "MARKETNAME", "SCADAVALUE"]

    # Combine scada data from interconnectors and dispatch units.
    scada_elements = pd.concat([scada, inter_flows], sort=False)

    # Merge the fcas and scada data based on time stamp, these leads every scada element to be joined to every fcas
    # element that then allows them to be comapred.
    profile_comp = pd.merge(
        fcas4s, scada_elements, "inner", left_on="TIMESTAMP", right_on="SETTLEMENTDATE"
    )

    # Calculate the error between each measurement.
    profile_comp["ERROR"] = profile_comp["VALUE"] - profile_comp["SCADAVALUE"]
    profile_comp["ERROR"] = profile_comp["ERROR"].abs()

    # Choose the fcas values that best matches the scada value during the 5 min interval.
    profile_comp = profile_comp.sort_values("ERROR")
    error_comp = profile_comp.groupby(
        ["MARKETNAME", "ELEMENTNUMBER", "TIMESTAMP"], as_index=False
    ).first()

    # Aggregate the error to comapre each scada and fcas element potential match.
    error_comp = error_comp.groupby(
        ["MARKETNAME", "ELEMENTNUMBER"], as_index=False
    ).sum()

    # Sort the comparisons based on aggregate error.
    error_comp = error_comp.sort_values("ERROR")

    # Drop duplicates of element numbers and scada element names, keeping the record for each with the least error.
    # Don't include units 0 values for scada
    best_matches_scada = error_comp[error_comp["SCADAVALUE"].abs() > 0]
    best_matches_scada = best_matches_scada.drop_duplicates(
        "ELEMENTNUMBER", keep="first"
    )
    best_matches_scada = best_matches_scada.drop_duplicates("MARKETNAME", keep="first")

    # Remove fcas elements where a match only occurred because both fcas and scada showed no dispatch.
    best_matches_scada["ELEMENTNUMBER"] = pd.to_numeric(
        best_matches_scada["ELEMENTNUMBER"]
    )
    best_matches_scada = best_matches_scada.sort_values("ELEMENTNUMBER")
    best_matches_scada["ELEMENTNUMBER"] = best_matches_scada["ELEMENTNUMBER"].astype(
        str
    )

    # Give error as a percentage.
    best_matches_scada["ERROR"] = (
        best_matches_scada["ERROR"] / best_matches_scada["SCADAVALUE"]
    )

    # drop matches with error greater than 100 %
    best_matches_scada = best_matches_scada[
        (best_matches_scada["ERROR"] < 1) & (best_matches_scada["ERROR"] > -1)
    ]

    best_matches_scada = best_matches_scada.loc[
        :, ("ELEMENTNUMBER", "MARKETNAME", "ERROR")
    ]

    if select_columns is not None:
        best_matches_scada = best_matches_scada.loc[:, select_columns]

    if filter_cols is not None:
        best_matches_scada = filters.filter_on_column_value(
            best_matches_scada, filter_cols, filter_values
        )

    return best_matches_scada


def capacity_factor(capacity_and_scada_grouped):
    scada_data = np.where(
        capacity_and_scada_grouped["SCADAVALUE"].isnull(),
        0.0,
        capacity_and_scada_grouped["SCADAVALUE"],
    )
    cf = scada_data / capacity_and_scada_grouped["MAXCAPACITY"]
    cf = cf.mean()
    return cf


def volume(capacity_and_scada_grouped):
    # Assumes 5 min Scada data
    return capacity_and_scada_grouped["SCADAVALUE"].sum() / 12


def volume_weighted_average_price(output, prices):
    v_by_p = output * prices
    total_cash = v_by_p.sum()
    total_volume = output.sum()
    vwap = total_cash / total_volume
    return vwap


def volume_weighted_average_trading_price(capacity_and_scada_grouped):
    return volume_weighted_average_price(
        capacity_and_scada_grouped["TRADING_TOTALCLEARED"],
        capacity_and_scada_grouped["TRADING_RRP"],
    )


def volume_weighted_average_spot_price(capacity_and_scada_grouped):
    return volume_weighted_average_price(
        capacity_and_scada_grouped["SCADAVALUE"],
        capacity_and_scada_grouped["DISPATCH_RRP"],
    )


def alt_volume_weighted_average_trading_price(capacity_and_scada_grouped):
    return volume_weighted_average_price(
        capacity_and_scada_grouped["TRADING_TOTALCLEARED"],
        capacity_and_scada_grouped["TRADING_VWAP"],
    )


def performance_at_nodal_peak(capacity_and_scada_grouped):
    index_max = capacity_and_scada_grouped["TOTALDEMAND"].idxmax()
    try:
        output_at_peak = capacity_and_scada_grouped["SCADAVALUE"][index_max]
    except:
        x = 1
    if np.isnan(output_at_peak):
        output_at_peak = 0
    performance = output_at_peak / capacity_and_scada_grouped["MAXCAPACITY"][index_max]
    return performance


def capacity_factor_over_90th_percentile_of_nodal_demand(capacity_and_scada_grouped):
    data_entries = len(capacity_and_scada_grouped["TOTALDEMAND"])
    enteries_in_90th_percentile = math.ceil(data_entries / 10)
    capacity_and_scada_grouped = capacity_and_scada_grouped.sort_values(
        "TOTALDEMAND", ascending=False
    )
    capacity_and_scada_grouped = capacity_and_scada_grouped.reset_index(drop=True)
    capacity_and_scada_grouped = capacity_and_scada_grouped.iloc[
        :enteries_in_90th_percentile, :
    ]
    cf = capacity_factor(capacity_and_scada_grouped)
    return cf


def stats_for_group(capacity_and_scada_grouped):
    cf = capacity_factor(capacity_and_scada_grouped)
    v = volume(capacity_and_scada_grouped)
    tvwap = volume_weighted_average_trading_price(capacity_and_scada_grouped)
    dvwap = volume_weighted_average_spot_price(capacity_and_scada_grouped)
    # alttvwap = alt_volume_weighted_average_trading_price(capacity_and_scada_grouped)
    peak = performance_at_nodal_peak(capacity_and_scada_grouped)
    peak_percentile = capacity_factor_over_90th_percentile_of_nodal_demand(
        capacity_and_scada_grouped
    )
    month = list(capacity_and_scada_grouped["MONTH"])[0]
    duid = list(capacity_and_scada_grouped["DUID"])[0]
    cf_df = pd.DataFrame(
        {
            "Month": [month],
            "DUID": [duid],
            "CapacityFactor": [cf],
            "Volume": [v],
            "TRADING_VWAP": [tvwap],
            "DISPATCH_VWAP": [dvwap],
            "NodalPeakCapacityFactor": peak,
            "Nodal90thPercentileCapacityFactor": [peak_percentile],
            # 'ALT_TRADING_VWAP': alttvwap
        }
    )
    return cf_df


def stats_by_month_and_plant(capacity_and_scada):
    # - timedelta(seconds=1)
    capacity_and_scada["effective_set_date"] = capacity_and_scada["SETTLEMENTDATE"]
    capacity_and_scada["MONTH"] = (
        capacity_and_scada["effective_set_date"].dt.year.astype(str)
        + "-"
        + capacity_and_scada["effective_set_date"].dt.month.astype(str).str.zfill(2)
    )
    capacity_factors = capacity_and_scada.groupby(
        ["MONTH", "DUID"], as_index=False
    ).apply(stats_for_group)
    return capacity_factors


def merge_tables_for_plant_stats(
    timeseries_df,
    gen_max_cap,
    gen_region,
    scada,
    trading_load,
    dispatch_price,
    trading_price,
    region_summary,
):
    gen_max_cap = gen_max_cap.sort_values("EFFECTIVEDATE")
    scada = scada.sort_values("SETTLEMENTDATE")
    merged_data = pd.merge_asof(
        scada,
        gen_max_cap,
        left_on="SETTLEMENTDATE",
        right_on="EFFECTIVEDATE",
        by="DUID",
    )
    gen_region = gen_region.sort_values("START_DATE")
    merged_data = pd.merge_asof(
        merged_data,
        gen_region,
        left_on="SETTLEMENTDATE",
        right_on="START_DATE",
        by="DUID",
    )
    merged_data = merged_data[~merged_data["REGIONID"].isnull()]
    merged_data = pd.merge(
        merged_data, trading_load, "left", on=["SETTLEMENTDATE", "DUID"]
    )
    merged_data = pd.merge(
        merged_data, dispatch_price, "left", on=["SETTLEMENTDATE", "REGIONID"]
    )
    merged_data = pd.merge(
        merged_data, trading_price, "left", on=["SETTLEMENTDATE", "REGIONID"]
    )
    merged_data = pd.merge(
        merged_data, region_summary, "left", on=["SETTLEMENTDATE", "REGIONID"]
    )
    merged_data = merged_data.loc[
        :,
        (
            "DUID",
            "EFFECTIVEDATE",
            "REGIONID",
            "SETTLEMENTDATE",
            "MAXCAPACITY",
            "SCADAVALUE",
            "TOTALCLEARED",
            "RRP_x",
            "RRP_y",
            "TOTALDEMAND",
        ),
    ]
    merged_data.columns = [
        "DUID",
        "DUDETAIL_EFFECTIVEDATE",
        "REGIONID",
        "SETTLEMENTDATE",
        "MAXCAPACITY",
        "SCADAVALUE",
        "TRADING_TOTALCLEARED",
        "DISPATCH_RRP",
        "TRADING_RRP",
        "TOTALDEMAND",
    ]
    return merged_data


def select_intervention_if_present(data, primary_key):
    data = data.sort_values(["INTERVENTION"])
    data = data.groupby(
        [col for col in primary_key if col != "INTERVENTION"], as_index=False
    ).last()
    return data


def select_highest_version_number(data, primary_key):
    data["VERSIONNO"] = pd.to_numeric(data["VERSIONNO"])
    data = data.sort_values(["VERSIONNO"])
    data["VERSIONNO"] = data["VERSIONNO"].astype(int).astype(str)
    data = data.groupby(
        [col for col in primary_key if col != "VERSIONNO"], as_index=False
    ).last()
    return data


def calc_trading_vwap_price(regional_demand, dispatch_price):
    demand_and_price = pd.merge(
        regional_demand, dispatch_price, "inner", on=["SETTLEMENTDATE", "REGIONID"]
    )
    demand_and_price["COST"] = demand_and_price["RRP"] * demand_and_price["TOTALDEMAND"]
    demand_and_price = demand_and_price.set_index("SETTLEMENTDATE")
    demand_and_price = (
        demand_and_price.groupby("REGIONID")
        .resample("30T", label="right", closed="right")
        .aggregate({"COST": "sum", "TOTALDEMAND": "sum"})
    )
    demand_and_price["TRADING_VWAP"] = (
        demand_and_price["COST"] / demand_and_price["TOTALDEMAND"]
    )
    demand_and_price.reset_index(inplace=True)
    demand_and_price = demand_and_price.loc[
        :, ("SETTLEMENTDATE", "REGIONID", "TRADING_VWAP")
    ]
    return demand_and_price


def calc_trading_load(scada):
    trading_load = scada.copy()
    trading_load = trading_load.set_index("SETTLEMENTDATE")
    trading_load["SCADAVALUE"] = pd.to_numeric(trading_load["SCADAVALUE"])
    trading_load = (
        trading_load.groupby("DUID")
        .resample("30T", label="right", closed="right")
        .aggregate({"SCADAVALUE": "mean"})
    )
    trading_load.reset_index(inplace=True)
    trading_load.columns = ["DUID", "SETTLEMENTDATE", "TOTALCLEARED"]
    return trading_load


def calc_trading_price(dispatch_price):
    dispatch_price = dispatch_price.set_index("SETTLEMENTDATE")
    dispatch_price = (
        dispatch_price.groupby("REGIONID")
        .resample("30T", label="right", closed="right")
        .aggregate({"RRP": "mean"})
    )
    dispatch_price.reset_index(inplace=True)
    return dispatch_price


def plant_stats(
    start_time,
    end_time,
    table_name,
    raw_data_location,
    select_columns=None,
    filter_cols=None,
    filter_values=None,
):

    ix = pd.date_range(
        start=datetime.strptime(start_time, "%Y/%m/%d %H:%M:%S"),
        end=datetime.strptime(end_time, "%Y/%m/%d %H:%M:%S") - timedelta(minutes=5),
        freq="5T",
    )
    timeseries_df = pd.DataFrame(index=ix)
    timeseries_df.reset_index(inplace=True)
    timeseries_df.columns = ["SETTLEMENTDATE"]

    gen_max_cap = data_fetch_methods.dynamic_data_compiler(
        start_time,
        end_time,
        "DUDETAIL",
        raw_data_location,
        select_columns=["EFFECTIVEDATE", "DUID", "VERSIONNO", "MAXCAPACITY"],
        filter_cols=filter_cols,
        filter_values=filter_values,
    )
    gen_max_cap = select_highest_version_number(
        gen_max_cap, defaults.table_primary_keys["DUDETAIL"]
    )
    gen_region = data_fetch_methods.dynamic_data_compiler(
        start_time,
        end_time,
        "DUDETAILSUMMARY",
        raw_data_location,
        select_columns=["START_DATE", "END_DATE", "DUID", "REGIONID"],
        filter_cols=filter_cols,
        filter_values=filter_values,
    )
    scada = data_fetch_methods.dynamic_data_compiler(
        start_time,
        end_time,
        "DISPATCH_UNIT_SCADA",
        raw_data_location,
        select_columns=["SETTLEMENTDATE", "DUID", "SCADAVALUE"],
    )
    dispatch_price = data_fetch_methods.dynamic_data_compiler(
        start_time,
        end_time,
        "DISPATCHPRICE",
        raw_data_location,
        select_columns=["SETTLEMENTDATE", "REGIONID", "RRP", "INTERVENTION"],
    )
    dispatch_price = select_intervention_if_present(
        dispatch_price, defaults.table_primary_keys["DISPATCHPRICE"]
    )
    trading_price = data_fetch_methods.dynamic_data_compiler(
        start_time,
        end_time,
        "TRADINGPRICE",
        raw_data_location,
        select_columns=["SETTLEMENTDATE", "REGIONID", "RRP"],
    )
    dispatch_price["RRP"] = pd.to_numeric(dispatch_price["RRP"])
    trading_price["RRP"] = pd.to_numeric(trading_price["RRP"])
    # trading_price = calc_trading_price(dispatch_price)

    region_summary = data_fetch_methods.dynamic_data_compiler(
        start_time,
        end_time,
        "DISPATCHREGIONSUM",
        raw_data_location,
        select_columns=[
            "SETTLEMENTDATE",
            "REGIONID",
            "TOTALDEMAND",
            "INTERVENTION",
            "DISPATCHINTERVAL",
        ],
    )
    region_summary = select_intervention_if_present(
        region_summary, defaults.table_primary_keys["DISPATCHREGIONSUM"]
    )

    scada_list = []
    for gen in scada.groupby(["DUID"], as_index=False):
        temp = pd.merge(timeseries_df, gen[1], "left", on="SETTLEMENTDATE")
        temp["SCADAVALUE"] = np.where(
            temp["SCADAVALUE"].isnull(), 0.0, temp["SCADAVALUE"]
        )
        temp["DUID"] = np.where(temp["DUID"].isnull(), gen[0], temp["DUID"])
        scada_list.append(temp)

    scada = pd.concat(scada_list)

    trading_load = calc_trading_load(scada)
    combined_data = merge_tables_for_plant_stats(
        timeseries_df,
        gen_max_cap,
        gen_region,
        scada,
        trading_load,
        dispatch_price,
        trading_price,
        region_summary,
    )
    combined_data["SCADAVALUE"] = pd.to_numeric(combined_data["SCADAVALUE"])
    combined_data["MAXCAPACITY"] = pd.to_numeric(combined_data["MAXCAPACITY"])
    combined_data["TRADING_TOTALCLEARED"] = pd.to_numeric(
        combined_data["TRADING_TOTALCLEARED"]
    )

    combined_data["TOTALDEMAND"] = pd.to_numeric(combined_data["TOTALDEMAND"])

    stats = stats_by_month_and_plant(combined_data)
    return stats


def trading_and_dispatch_cost():
    gen_region = data_fetch_methods.dynamic_data_compiler(
        "2017/01/01 00:05:00",
        "2018/01/01 00:05:00",
        "DUDETAILSUMMARY",
        defaults.raw_data_cache,
        select_columns=["START_DATE", "END_DATE", "DUID", "REGIONID"],
    )
    scada = data_fetch_methods.dynamic_data_compiler(
        "2017/01/01 00:05:00",
        "2018/01/01 00:05:00",
        "DISPATCH_UNIT_SCADA",
        defaults.raw_data_cache,
    )

    ix = pd.date_range(
        start=datetime.strptime("2017/01/01 00:00:00", "%Y/%m/%d %H:%M:%S"),
        end=datetime.strptime("2018/01/01 00:00:00", "%Y/%m/%d %H:%M:%S"),
        freq="5T",
    )
    timeseries_df = pd.DataFrame(index=ix)
    timeseries_df.reset_index(inplace=True)
    timeseries_df.columns = ["SETTLEMENTDATE"]

    scada_list = []
    for gen in scada.groupby(["DUID"], as_index=False):
        temp = pd.merge(timeseries_df, gen[1], "left", on="SETTLEMENTDATE")
        temp["SCADAVALUE"] = np.where(
            temp["SCADAVALUE"].isnull(), 0.0, temp["SCADAVALUE"]
        )
        temp["DUID"] = np.where(temp["DUID"].isnull(), gen[0], temp["DUID"])
        scada_list.append(temp)

    scada = pd.concat(scada_list)

    dispatch_price = data_fetch_methods.dynamic_data_compiler(
        "2017/01/01 00:00:00",
        "2018/01/01 00:05:00",
        "DISPATCHPRICE",
        defaults.raw_data_cache,
    )
    dispatch_price = select_intervention_if_present(
        dispatch_price, defaults.table_primary_keys["DISPATCHPRICE"]
    )
    gen_region = gen_region.sort_values("START_DATE")
    scada = scada.sort_values("SETTLEMENTDATE")
    scada_and_regions = pd.merge_asof(
        scada, gen_region, left_on="SETTLEMENTDATE", right_on="START_DATE", by="DUID"
    )
    scada_and_regions = scada_and_regions[~scada_and_regions["REGIONID"].isnull()]
    scada_and_regions = pd.merge(
        scada_and_regions,
        dispatch_price,
        "inner",
        left_on=["REGIONID", "SETTLEMENTDATE"],
        right_on=["REGIONID", "SETTLEMENTDATE"],
    )

    scada_and_regions["SCADAVALUE"] = pd.to_numeric(scada_and_regions["SCADAVALUE"])
    scada_and_regions["RRP"] = pd.to_numeric(scada_and_regions["RRP"])

    scada_and_regions["DISPATCHCOST"] = (
        scada_and_regions["SCADAVALUE"] * scada_and_regions["RRP"]
    ) / 12

    scada_and_regions = scada_and_regions.set_index("SETTLEMENTDATE")
    scada_and_regions = (
        scada_and_regions.groupby("DUID")
        .resample("30T", label="right", closed="right")
        .aggregate({"SCADAVALUE": "mean", "RRP": "mean", "DISPATCHCOST": "sum"})
    )
    scada_and_regions.reset_index(inplace=True)

    scada_and_regions["TRADINGCOST"] = (
        scada_and_regions["SCADAVALUE"] * scada_and_regions["RRP"]
    ) / 2

    scada_and_regions["SCADAVALUE"] = scada_and_regions["SCADAVALUE"] / 2

    scada_and_regions = scada_and_regions.groupby("DUID").sum()

    scada_and_regions.to_csv("C:/Users/user/Documents/dispatch_trading_cost.csv")



import unittest

from src.nemosis import processing_info_maps, defaults, query_wrappers, data_fetch_methods
import pandas as pd
from datetime import datetime, timedelta


class TestSearchTypeValidity(unittest.TestCase):
    def setUp(self):
        self.time_yesterday = (datetime.now() - timedelta(days=1)).replace(
            hour=0, minute=0, second=0, microsecond=0
        )
        pass

    def test_start_to_end_no_duplication_between_batches(self):
        for table_name in processing_info_maps.search_type.keys():
            if processing_info_maps.search_type[table_name] == "start_to_end":
                print("Validating start_to_end type for table {}".format(table_name))
                start_time = datetime.strptime(
                    "2018/06/01 00:00:00", "%Y/%m/%d %H:%M:%S"
                )
                end_time = datetime.strptime("2018/09/01 00:00:00", "%Y/%m/%d %H:%M:%S")
                if table_name in ["DAILY_REGION_SUMMARY", "NEXT_DAY_DISPATCHLOAD", "INTERMITTENT_GEN_SCADA"]:
                    end_time = self.time_yesterday
                    start_time = self.time_yesterday - timedelta(days=8)
                if table_name in ["FCAS_4_SECOND"]:
                    end_time = self.time_yesterday
                    start_time = self.time_yesterday - timedelta(hours=2)
                data_tables = data_fetch_methods._dynamic_data_fetch_loop(
                    start_search=start_time,
                    start_time=start_time,
                    end_time=end_time,
                    table_name=table_name,
                    raw_data_location=defaults.raw_data_cache,
                    select_columns=defaults.table_primary_keys[table_name],
                    date_filter=None,
                    keep_csv=False,
                )
                all_data = pd.concat(data_tables, sort=False)
                contains_duplicates = all_data.duplicated().any()
                self.assertEqual(
                    False, contains_duplicates, "table {}".format(table_name)
                )
                print("Type valid, no duplicates found.")

    def test_start_to_end_has_settlement_or_interval_col(self):
        for table_name in processing_info_maps.search_type.keys():
            if processing_info_maps.search_type[table_name] == "start_to_end":
                has_settlement_date_col = (
                    "SETTLEMENTDATE" in defaults.table_columns[table_name]
                )
                has_interval_datetime_col = (
                    "INTERVAL_DATETIME" in defaults.table_columns[table_name]
                )
                has_interval_timestamp_col = (
                    "TIMESTAMP" in defaults.table_columns[table_name]
                )
                has_run_datetime_col = (
                    "RUN_DATETIME" in defaults.table_columns[table_name]
                )
                has_either = (
                    has_interval_datetime_col
                    or has_settlement_date_col
                    or has_interval_timestamp_col
                    or has_run_datetime_col
                )
                self.assertEqual(True, has_either)
                print(
                    "{} is valid candidate for type start_to_end as there is a SETTLEMENTDATE, "
                    "INTERVAL_DATETIME or TIMESTAMP column to filter on.".format(
                        table_name
                    )
                )

    def test_all_no_duplication_between_batches(self):
        for table_name in processing_info_maps.search_type.keys():
            if processing_info_maps.search_type[table_name] == "all":
                print("Validating all type for table {}".format(table_name))
                if table_name in [
                    "GENCONDATA",
                    "SPDCONNECTIONPOINTCONSTRAINT",
                    "SPDINTERCONNECTORCONSTRAINT",
                    "DUDETAILSUMMARY",
                    "LOSSMODEL",
                    "LOSSFACTORMODEL",
                    "MNSP_DAYOFFER",
                    "MNSP_PEROFFER",
                    "MNSP_INTERCONNECTOR",
                    "INTERCONNECTOR",
                    "INTERCONNECTORCONSTRAINT",
                    "DUDETAIL",
                    "MARKET_PRICE_THRESHOLDS",
                    "PARTICIPANT"
                ]:
                    print(
                        "{} is known to contain duplicate entries and is exempted from this test, a finalise "
                        "data processing step is included in dynamic data fetch to clean up these duplicates.".format(
                            table_name
                        )
                    )
                    continue
                start_test_window = defaults.nem_data_model_start_time
                start_time = datetime.strptime(start_test_window, "%Y/%m/%d %H:%M:%S")
                end_time = datetime.strptime("2018/01/01 00:00:00", "%Y/%m/%d %H:%M:%S")
                start_search = datetime.strptime(start_test_window, "%Y/%m/%d %H:%M:%S")
                data_tables = data_fetch_methods._dynamic_data_fetch_loop(
                    start_search=start_search,
                    start_time=start_time,
                    end_time=end_time,
                    table_name=table_name,
                    raw_data_location=defaults.raw_data_cache,
                    select_columns=defaults.table_primary_keys[table_name],
                    date_filter=None,
                    keep_csv=False,
                )
                all_data = pd.concat(data_tables, sort=False)
                contains_duplicates = all_data.duplicated().any()
                self.assertEqual(
                    False, contains_duplicates, "table {}".format(table_name)
                )
                print("Type valid, no duplicates found.")

    def test_all_no_duplication_between_batches_with_finalise_step(self):
        for table_name in processing_info_maps.search_type.keys():
            if processing_info_maps.search_type[table_name] == "all":
                print("Testing duplicate removal for table {}".format(table_name))
                start_test_window = defaults.nem_data_model_start_time
                # start_test_window = '2018/01/01 00:00:00'
                start_time = datetime.strptime(start_test_window, "%Y/%m/%d %H:%M:%S")
                end_time = datetime.strptime("2018/01/01 00:00:00", "%Y/%m/%d %H:%M:%S")
                start_search = datetime.strptime(start_test_window, "%Y/%m/%d %H:%M:%S")
                data_tables = data_fetch_methods._dynamic_data_fetch_loop(
                    start_search=start_search,
                    start_time=start_time,
                    end_time=end_time,
                    table_name=table_name,
                    raw_data_location=defaults.raw_data_cache,
                    select_columns=defaults.table_primary_keys[table_name],
                    date_filter=None,
                    keep_csv=False,
                )
                all_data = pd.concat(data_tables, sort=False)
                all_data = query_wrappers.drop_duplicates_by_primary_key(
                    all_data, start_time, table_name
                )
                contains_duplicates = all_data.duplicated().any()
                self.assertEqual(False, contains_duplicates)
                print("Type valid, no duplicates found.")

    def test_start_to_end_has_no_settlement_interval_or_timestamp_col(self):
        for table_name in processing_info_maps.search_type.keys():
            if processing_info_maps.search_type[table_name] == "all":
                has_settlement_date_col = (
                    "SETTLEMENTDATE" in defaults.table_columns[table_name]
                )
                has_interval_datetime_col = (
                    "INTERVAL_DATETIME" in defaults.table_columns[table_name]
                )
                has_interval_timestamp_col = (
                    "TIMESTAMP" in defaults.table_columns[table_name]
                )
                has_either = (
                    has_interval_datetime_col
                    or has_settlement_date_col
                    or has_interval_timestamp_col
                )
                self.assertEqual(False, has_either, "table {}".format(table_name))
                print(
                    "{} is valid candidate for type all as there is not a SETTLEMENTDATE, "
                    "INTERVAL_DATETIME or TIMESTAMP column to filter on".format(
                        table_name
                    )
                )

    def test_last_contains_data_from_first(self):
        for table_name in processing_info_maps.search_type.keys():
            if processing_info_maps.search_type[table_name] == "end":
                start_test_window = defaults.nem_data_model_start_time
                # start_test_window = '2018/01/01 00:00:00'
                start_time = datetime.strptime(start_test_window, "%Y/%m/%d %H:%M:%S")
                end_time = datetime.strptime("2018/01/01 00:00:00", "%Y/%m/%d %H:%M:%S")
                start_search = datetime.strptime(start_test_window, "%Y/%m/%d %H:%M:%S")
                select_columns = None
                (
                    _,
                    _,
                    select_columns,
                    _,
                    _,
                ) = data_fetch_methods._set_up_dynamic_compilers(
                    table_name, start_time, end_time, select_columns
                )
                data_tables = data_fetch_methods._dynamic_data_fetch_loop(
                    start_search=start_search,
                    start_time=start_time,
                    end_time=end_time,
                    table_name=table_name,
                    raw_data_location=defaults.raw_data_cache,
                    select_columns=select_columns,
                    date_filter=None,
                    keep_csv=False,
                )
                first_data_table = data_tables[35].loc[
                    :, defaults.table_primary_keys[table_name]
                                   ]
                last_data_table = data_tables[-1]
                comp = pd.merge(
                    first_data_table,
                    last_data_table,
                    "left",
                    defaults.table_primary_keys[table_name],
                )
                non_primary_col = [
                    col
                    for col in defaults.table_columns[table_name]
                    if col not in defaults.table_primary_keys[table_name]
                ][0]
                missing_from_last = comp[comp[non_primary_col].isnull()]
                self.assertEqual(False, missing_from_last.empty)



import unittest
import pandas as pd
from src.nemosis import filters
from pandas.testing import assert_frame_equal
from datetime import datetime
import numpy as np


class TestFiltersStartDate(unittest.TestCase):
    def setUp(self):
        self.start_date_data = pd.DataFrame(
            {
                "START_DATE": ["2011/01/01 00:00:00", "2015/01/01 00:00:00"],
                "END_DATE": ["2015/01/01 00:00:00", "2015/07/01 00:12:00"],
            }
        )

    def test_start_date_pick_first_of_two(self):
        start_time = datetime.strptime("2014/06/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2014/09/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_start_and_end_date(
            self.start_date_data, start_time=start_time, end_time=end_time
        )
        aim = pd.DataFrame(
            {"START_DATE": ["2011/01/01 00:00:00"], "END_DATE": ["2015/01/01 00:00:00"]}
        )
        aim["START_DATE"] = pd.to_datetime(
            aim["START_DATE"], format="%Y/%m/%d %H:%M:%S"
        )
        aim["END_DATE"] = pd.to_datetime(aim["END_DATE"], format="%Y/%m/%d %H:%M:%S")
        assert_frame_equal(aim, result)

    def test_start_date_pick_second_of_two(self):
        start_time = datetime.strptime("2015/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2015/05/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_start_and_end_date(
            self.start_date_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame(
            {"START_DATE": ["2015/01/01 00:00:00"], "END_DATE": ["2015/07/01 00:12:00"]}
        )
        aim["START_DATE"] = pd.to_datetime(
            aim["START_DATE"], format="%Y/%m/%d %H:%M:%S"
        )
        aim["END_DATE"] = pd.to_datetime(aim["END_DATE"], format="%Y/%m/%d %H:%M:%S")
        assert_frame_equal(aim, result)

    def test_start_date_pick_two_of_two_by_overlaping_interval(self):
        start_time = datetime.strptime("2011/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2015/05/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_start_and_end_date(
            self.start_date_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame(
            {
                "START_DATE": ["2011/01/01 00:00:00", "2015/01/01 00:00:00"],
                "END_DATE": ["2015/01/01 00:00:00", "2015/07/01 00:12:00"],
            }
        )
        aim["START_DATE"] = pd.to_datetime(
            aim["START_DATE"], format="%Y/%m/%d %H:%M:%S"
        )
        aim["END_DATE"] = pd.to_datetime(aim["END_DATE"], format="%Y/%m/%d %H:%M:%S")
        assert_frame_equal(aim, result)

    def test_start_date_pick_none_of_two_by_overshooting_date(self):
        start_time = datetime.strptime("2018/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2019/05/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_start_and_end_date(
            self.start_date_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame({"START_DATE": [], "END_DATE": []})
        aim["START_DATE"] = pd.to_datetime(
            aim["START_DATE"], format="%Y/%m/%d %H:%M:%S"
        )
        aim["END_DATE"] = pd.to_datetime(aim["END_DATE"], format="%Y/%m/%d %H:%M:%S")
        assert_frame_equal(aim, result)

    def test_start_date_pick_none_of_two_by_undershooting_date(self):
        start_time = datetime.strptime("2010/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2010/05/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_start_and_end_date(
            self.start_date_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame({"START_DATE": [], "END_DATE": []})
        aim["START_DATE"] = pd.to_datetime(
            aim["START_DATE"], format="%Y/%m/%d %H:%M:%S"
        )
        aim["END_DATE"] = pd.to_datetime(aim["END_DATE"], format="%Y/%m/%d %H:%M:%S")
        assert_frame_equal(aim, result)


class TestFiltersEffectiveDate(unittest.TestCase):
    def setUp(self):
        self.last_changed_data = pd.DataFrame(
            {"EFFECTIVEDATE": ["2011/01/01 00:00:00", "2015/01/01 00:00:00"]}
        )

    def test_start_date_pick_first_of_two(self):
        start_time = datetime.strptime("2014/06/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2014/09/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_effective_date(
            self.last_changed_data, start_time=start_time, end_time=end_time
        )
        aim = pd.DataFrame({"EFFECTIVEDATE": ["2011/01/01 00:00:00"]})
        aim["EFFECTIVEDATE"] = pd.to_datetime(
            aim["EFFECTIVEDATE"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)

    def test_start_date_two_of_two_with_date_window_on_second(self):
        start_time = datetime.strptime("2015/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2015/05/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_effective_date(
            self.last_changed_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame(
            {"EFFECTIVEDATE": ["2011/01/01 00:00:00", "2015/01/01 00:00:00"]}
        )
        aim["EFFECTIVEDATE"] = pd.to_datetime(
            aim["EFFECTIVEDATE"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)

    def test_start_date_pick_two_of_two_by_overlaping_interval(self):
        start_time = datetime.strptime("2011/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2015/05/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_effective_date(
            self.last_changed_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame(
            {"EFFECTIVEDATE": ["2011/01/01 00:00:00", "2015/01/01 00:00:00"]}
        )
        aim["EFFECTIVEDATE"] = pd.to_datetime(
            aim["EFFECTIVEDATE"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)

    def test_start_date_pick_two_of_two_by_overshooting_date(self):
        start_time = datetime.strptime("2018/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2019/05/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_effective_date(
            self.last_changed_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame(
            {"EFFECTIVEDATE": ["2011/01/01 00:00:00", "2015/01/01 00:00:00"]}
        )
        aim["EFFECTIVEDATE"] = pd.to_datetime(
            aim["EFFECTIVEDATE"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)

    def test_start_date_pick_none_of_two_by_undershooting_date(self):
        start_time = datetime.strptime("2010/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2010/05/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_effective_date(
            self.last_changed_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame({"EFFECTIVEDATE": []})
        aim["EFFECTIVEDATE"] = pd.to_datetime(
            aim["EFFECTIVEDATE"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)


class TestFiltersSettlementDate(unittest.TestCase):
    def setUp(self):
        self.settlement_date_data = pd.DataFrame(
            {"SETTLEMENTDATE": ["2011/01/01 00:00:00", "2015/01/01 00:00:00"]}
        )

    def test_settlement_date_pick_first_of_two(self):
        start_time = datetime.strptime("2010/06/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2011/09/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_settlementdate(
            self.settlement_date_data, start_time=start_time, end_time=end_time
        )
        aim = pd.DataFrame({"SETTLEMENTDATE": ["2011/01/01 00:00:00"]})
        aim["SETTLEMENTDATE"] = pd.to_datetime(
            aim["SETTLEMENTDATE"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)

    def test_settlement_date_pick_second_of_two(self):
        start_time = datetime.strptime("2014/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2015/05/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_settlementdate(
            self.settlement_date_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame({"SETTLEMENTDATE": ["2015/01/01 00:00:00"]})
        aim["SETTLEMENTDATE"] = pd.to_datetime(
            aim["SETTLEMENTDATE"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)

    def test_settlement_date_pick_two_of_two_by_overlaping_interval(self):
        start_time = datetime.strptime("2010/12/31 23:59:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2015/05/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        aim = pd.DataFrame(
            {"SETTLEMENTDATE": ["2011/01/01 00:00:00", "2015/01/01 00:00:00"]}
        )
        result = filters.filter_on_settlementdate(
            self.settlement_date_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim["SETTLEMENTDATE"] = pd.to_datetime(
            aim["SETTLEMENTDATE"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)

    def test_settlement_date_pick_none_of_two_by_overshooting_date(self):
        start_time = datetime.strptime("2018/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2019/05/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_settlementdate(
            self.settlement_date_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame({"SETTLEMENTDATE": []})
        aim["SETTLEMENTDATE"] = pd.to_datetime(
            aim["SETTLEMENTDATE"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)

    def test_settlement_date_pick_none_of_two_by_undershooting_date(self):
        start_time = datetime.strptime("2010/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2010/05/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_settlementdate(
            self.settlement_date_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame({"SETTLEMENTDATE": []})
        aim["SETTLEMENTDATE"] = pd.to_datetime(
            aim["SETTLEMENTDATE"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)

    def test_settlement_date_end_date_exclusive_by_undershooting(self):
        start_time = datetime.strptime("2010/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2010/12/31 23:59:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_settlementdate(
            self.settlement_date_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame({"SETTLEMENTDATE": []})
        aim["SETTLEMENTDATE"] = pd.to_datetime(
            aim["SETTLEMENTDATE"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)


class TestFiltersTimeStamp(unittest.TestCase):
    def setUp(self):
        self.time_stamp_data = pd.DataFrame(
            {"TIMESTAMP": ["2011/01/01 00:00:00", "2015/01/01 00:00:00"]}
        )

    def test_time_stamp_pick_first_of_two(self):
        start_time = datetime.strptime("2010/06/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2011/09/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_timestamp(
            self.time_stamp_data, start_time=start_time, end_time=end_time
        )
        aim = pd.DataFrame({"TIMESTAMP": ["2011/01/01 00:00:00"]})
        aim["TIMESTAMP"] = pd.to_datetime(aim["TIMESTAMP"], format="%Y/%m/%d %H:%M:%S")
        assert_frame_equal(aim, result)

    def test_time_stamp_pick_second_of_two(self):
        start_time = datetime.strptime("2014/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2015/05/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_timestamp(
            self.time_stamp_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame({"TIMESTAMP": ["2015/01/01 00:00:00"]})
        aim["TIMESTAMP"] = pd.to_datetime(aim["TIMESTAMP"], format="%Y/%m/%d %H:%M:%S")
        assert_frame_equal(aim, result)

    def test_time_stamp_pick_two_of_two_by_overlaping_interval(self):
        start_time = datetime.strptime("2010/12/31 23:59:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2015/05/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        aim = pd.DataFrame(
            {"TIMESTAMP": ["2011/01/01 00:00:00", "2015/01/01 00:00:00"]}
        )
        result = filters.filter_on_timestamp(
            self.time_stamp_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim["TIMESTAMP"] = pd.to_datetime(aim["TIMESTAMP"], format="%Y/%m/%d %H:%M:%S")
        assert_frame_equal(aim, result)

    def test_time_stamp_pick_none_of_two_by_overshooting_date(self):
        start_time = datetime.strptime("2018/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2019/05/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_timestamp(
            self.time_stamp_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame({"TIMESTAMP": []})
        aim["TIMESTAMP"] = pd.to_datetime(aim["TIMESTAMP"], format="%Y/%m/%d %H:%M:%S")
        assert_frame_equal(aim, result)

    def test_time_stamp_pick_none_of_two_by_undershooting_date(self):
        start_time = datetime.strptime("2010/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2010/05/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_timestamp(
            self.time_stamp_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame({"TIMESTAMP": []})
        aim["TIMESTAMP"] = pd.to_datetime(aim["TIMESTAMP"], format="%Y/%m/%d %H:%M:%S")
        assert_frame_equal(aim, result)

    def test_time_stamp_end_date_exclusive_by_undershooting(self):
        start_time = datetime.strptime("2010/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2010/12/31 23:59:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_timestamp(
            self.time_stamp_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame({"TIMESTAMP": []})
        aim["TIMESTAMP"] = pd.to_datetime(aim["TIMESTAMP"], format="%Y/%m/%d %H:%M:%S")
        assert_frame_equal(aim, result)


class TestFiltersIntervalDatetime(unittest.TestCase):
    def setUp(self):
        self.interval_datetime_data = pd.DataFrame(
            {"INTERVAL_DATETIME": ["2011/01/01 00:00:00", "2015/01/01 00:00:00"]}
        )

    def test_interval_datetime_pick_first_of_two(self):
        start_time = datetime.strptime("2010/06/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2011/09/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_interval_datetime(
            self.interval_datetime_data, start_time=start_time, end_time=end_time
        )
        aim = pd.DataFrame({"INTERVAL_DATETIME": ["2011/01/01 00:00:00"]})
        aim["INTERVAL_DATETIME"] = pd.to_datetime(
            aim["INTERVAL_DATETIME"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)

    def test_interval_datetime_pick_second_of_two(self):
        start_time = datetime.strptime("2014/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2015/05/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_interval_datetime(
            self.interval_datetime_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame({"INTERVAL_DATETIME": ["2015/01/01 00:00:00"]})
        aim["INTERVAL_DATETIME"] = pd.to_datetime(
            aim["INTERVAL_DATETIME"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)

    def test_interval_datetime_pick_two_of_two_by_overlaping_interval(self):
        start_time = datetime.strptime("2010/12/31 23:59:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2015/05/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        aim = pd.DataFrame(
            {"INTERVAL_DATETIME": ["2011/01/01 00:00:00", "2015/01/01 00:00:00"]}
        )
        result = filters.filter_on_interval_datetime(
            self.interval_datetime_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim["INTERVAL_DATETIME"] = pd.to_datetime(
            aim["INTERVAL_DATETIME"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)

    def test_interval_datetime_pick_none_of_two_by_overshooting_date(self):
        start_time = datetime.strptime("2018/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2019/05/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_interval_datetime(
            self.interval_datetime_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame({"INTERVAL_DATETIME": []})
        aim["INTERVAL_DATETIME"] = pd.to_datetime(
            aim["INTERVAL_DATETIME"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)

    def test_interval_datetime_stamp_pick_none_of_two_by_undershooting_date(self):
        start_time = datetime.strptime("2010/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2010/05/01 00:13:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_interval_datetime(
            self.interval_datetime_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame({"INTERVAL_DATETIME": []})
        aim["INTERVAL_DATETIME"] = pd.to_datetime(
            aim["INTERVAL_DATETIME"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)

    def test_interval_datetime_end_date_exclusive_by_undershooting(self):
        start_time = datetime.strptime("2010/04/01 00:00:00", "%Y/%m/%d %H:%M:%S")
        end_time = datetime.strptime("2010/12/31 23:59:00", "%Y/%m/%d %H:%M:%S")
        result = filters.filter_on_interval_datetime(
            self.interval_datetime_data, start_time=start_time, end_time=end_time
        ).reset_index(drop=True)
        aim = pd.DataFrame({"INTERVAL_DATETIME": []})
        aim["INTERVAL_DATETIME"] = pd.to_datetime(
            aim["INTERVAL_DATETIME"], format="%Y/%m/%d %H:%M:%S"
        )
        assert_frame_equal(aim, result)


class TestFiltersColumnValue(unittest.TestCase):
    def setUp(self):
        self.int_filter_data = pd.DataFrame({"INTCOL": [1, 10, 100, -5, 0, 10, 10]})
        self.int_and_string_filter_data = pd.DataFrame(
            {
                "INTCOL": [1, 10, 100, -5, 0, 10, 10],
                "STRINGCOL": ["1", "10", "100", "-5", "0", "10", "10"],
            }
        )

    def test_filter_one_col_one_value_positive_ints(self):
        filter_cols = ("INTCOL",)
        filter_values = ([10],)
        result = filters.filter_on_column_value(
            self.int_filter_data, filter_cols, filter_values
        ).reset_index(drop=True)
        aim = pd.DataFrame({"INTCOL": [10, 10, 10]})
        assert_frame_equal(aim, result)

    def test_filter_one_col_two_values_positive_and_negative_ints(self):
        filter_cols = ("INTCOL",)
        filter_values = ([10, -5],)
        result = filters.filter_on_column_value(
            self.int_filter_data, filter_cols, filter_values
        ).reset_index(drop=True)
        aim = pd.DataFrame({"INTCOL": [10, -5, 10, 10]})
        assert_frame_equal(aim, result)

    def test_filter_two_cols_one_value_each_not_matching(self):
        filter_cols = ("INTCOL", "STRINGCOL")
        filter_values = ([10], ["100"])
        result = filters.filter_on_column_value(
            self.int_and_string_filter_data, filter_cols, filter_values
        ).reset_index(drop=True)
        aim = pd.DataFrame({"INTCOL": [], "STRINGCOL": []})
        aim = aim.astype(dtype={"INTCOL": np.int64, "STRINGCOL": str})
        assert_frame_equal(aim, result)

    def test_filter_two_cols_one_value_each_matching(self):
        filter_cols = ("INTCOL", "STRINGCOL")
        filter_values = ([10], ["10"])
        result = filters.filter_on_column_value(
            self.int_and_string_filter_data, filter_cols, filter_values
        ).reset_index(drop=True)
        aim = pd.DataFrame({"INTCOL": [10, 10, 10], "STRINGCOL": ["10", "10", "10"]})
        assert_frame_equal(aim, result)

    def test_filter_just_one_of_two_cols(self):
        filter_cols = ("INTCOL",)
        filter_values = ([10],)
        result = filters.filter_on_column_value(
            self.int_and_string_filter_data, filter_cols, filter_values
        ).reset_index(drop=True)
        aim = pd.DataFrame({"INTCOL": [10, 10, 10], "STRINGCOL": ["10", "10", "10"]})
        assert_frame_equal(aim, result)

    def test_filter_one_empty_values_returns_empty_data_frame(self):
        filter_cols = ("INTCOL",)
        filter_values = ([],)
        result = filters.filter_on_column_value(
            self.int_and_string_filter_data, filter_cols, filter_values
        ).reset_index(drop=True)
        aim = pd.DataFrame({"INTCOL": [], "STRINGCOL": []})
        aim = aim.astype(dtype={"INTCOL": np.int64, "STRINGCOL": str})
        assert_frame_equal(aim, result)


if __name__ == "__main__":
    unittest.main()



import os
import logging
import requests
from bs4 import BeautifulSoup
import zipfile
import io
import pandas as pd

from . import defaults, custom_errors

logger = logging.getLogger(__name__)

# Windows Chrome for User-Agent request headers
USR_AGENT_HEADER = {
    "User-Agent": (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64)"
        + " AppleWebKit/537.36 (KHTML, like Gecko) "
        + "Chrome/80.0.3987.87 Safari/537.36"
    )
}


def run(year, month, day, chunk, index, filename_stub, down_load_to):
    """This function"""

    url = defaults.aemo_mms_url
    # Add the year and month information to the generic AEMO data url
    url_formatted = format_aemo_url(url, year, month, filename_stub)

    # Perform the download, unzipping saving of the file
    try:
        download_unzip_csv(url_formatted, down_load_to)
    except Exception:
        if chunk == 1:
            logger.warning(f"{filename_stub} not downloaded")


def run_bid_tables(year, month, day, chunk, index, filename_stub, down_load_to):
    if day is None:
        run(year, month, day, chunk, index, filename_stub, down_load_to)
    else:
        try:
            filename_stub = "BIDMOVE_COMPLETE_{year}{month}{day}".format(year=year, month=month, day=day)
            download_url = _get_current_url(
                filename_stub,
                defaults.current_data_page_urls["BIDDING"])
            _download_and_unpack_bid_move_complete_files(
                download_url, down_load_to
            )
        except Exception:
            logger.warning(f"{filename_stub} not downloaded")


def run_next_day_region_tables(year, month, day, chunk, index, filename_stub, down_load_to):
    try:
        filename_stub = "PUBLIC_DAILY_{year}{month}{day}".format(year=year, month=month, day=day)
        download_url = _get_current_url(
            filename_stub,
            defaults.current_data_page_urls["DAILY_REGION_SUMMARY"])
        _download_and_unpack_next_region_tables(
            download_url, down_load_to
        )
    except Exception:
        logger.warning(f"{filename_stub} not downloaded")


def run_next_dispatch_tables(year, month, day, chunk, index, filename_stub, down_load_to):
    try:
        filename_stub = "PUBLIC_NEXT_DAY_DISPATCH_{year}{month}{day}".format(year=year, month=month, day=day)
        download_url = _get_current_url(
            filename_stub,
            defaults.current_data_page_urls["NEXT_DAY_DISPATCHLOAD"])
        _download_and_unpack_next_dispatch_load_files_complete_files(download_url, down_load_to)
    except Exception:
        logger.warning(f"{filename_stub} not downloaded")


def run_intermittent_gen_scada(year, month, day, chunk, index, filename_stub, down_load_to):
    try:
        download_url = _get_current_url(
            filename_stub,
            defaults.current_data_page_urls["INTERMITTENT_GEN_SCADA"])
        _download_and_unpack_intermittent_gen_scada_file(download_url, down_load_to)
    except Exception:
        logger.warning(f"{filename_stub} not downloaded")


def _get_current_url(filename_stub, current_page_url):
    sub_url = _get_matching_link(
        url=defaults.nem_web_domain_url + current_page_url,
        stub_link=filename_stub)
    return defaults.nem_web_domain_url + sub_url


def _download_and_unpack_bid_move_complete_files(
    download_url, down_load_to
):
    r = requests.get(download_url, headers=USR_AGENT_HEADER)
    zipped_file = zipfile.ZipFile(io.BytesIO(r.content))

    file_name = zipped_file.namelist()[
        0
    ]  # Just one file so we can pull it out of the list using 0
    start_row_second_table = _find_start_row_nth_table(
        zipped_file, file_name, 2
    )
    csv_file = zipped_file.open(file_name)
    BIDDAYOFFER_D = pd.read_csv(
        csv_file, header=1, nrows=start_row_second_table - 3, dtype=str
    )
    BIDDAYOFFER_D.to_csv(
        os.path.join(
            down_load_to,
            "PUBLIC_DVD_BIDDAYOFFER_D_" + file_name[24:32] + ".csv",
        ),
        index=False,
    )
    csv_file = zipped_file.open(file_name)
    BIDPEROFFER_D = pd.read_csv(
        csv_file, header=start_row_second_table - 1, dtype=str
    )[:-1]
    BIDPEROFFER_D.to_csv(
        os.path.join(
            down_load_to,
            "PUBLIC_DVD_BIDPEROFFER_D_" + file_name[24:32] + ".csv",
        ),
        index=False,
    )


def _download_and_unpack_next_region_tables(
    download_url, down_load_to
):
    r = requests.get(download_url, headers=USR_AGENT_HEADER)
    zipped_file = zipfile.ZipFile(io.BytesIO(r.content))

    file_name = zipped_file.namelist()[
        0
    ]  # Just one file so we can pull it out of the list using 0
    start_row_second_table = _find_start_row_nth_table(
        zipped_file, file_name, 2
    )
    start_row_third_table = _find_start_row_nth_table(
        zipped_file, file_name, 3
    )
    csv_file = zipped_file.open(file_name)
    DAILY_REGION_SUMMARY = pd.read_csv(
        csv_file, header=start_row_second_table - 1,
        nrows=start_row_third_table - start_row_second_table - 1, dtype=str
    )
    DAILY_REGION_SUMMARY.to_csv(
        os.path.join(
            down_load_to,
            "PUBLIC_DAILY_REGION_SUMMARY_" + file_name[13:21] + ".csv",
        ),
        index=False,
    )
    

def _download_and_unpack_next_dispatch_load_files_complete_files(
    download_url, down_load_to
):
    r = requests.get(download_url, headers=USR_AGENT_HEADER)
    zipped_file = zipfile.ZipFile(io.BytesIO(r.content))

    file_name = zipped_file.namelist()[
        0
    ]  # Just one file so we can pull it out of the list using 0
    start_row_second_table = _find_start_row_nth_table(
        zipped_file, file_name, 2
    )
    csv_file = zipped_file.open(file_name)
    NEXT_DAY_DISPATCHLOAD = pd.read_csv(
        csv_file, header=1, nrows=start_row_second_table - 3, dtype=str
    )
    NEXT_DAY_DISPATCHLOAD.to_csv(
        os.path.join(
            down_load_to,
            "PUBLIC_NEXT_DAY_DISPATCHLOAD_" + file_name[25:33] + ".csv",
        ),
        index=False,
    )


def _download_and_unpack_intermittent_gen_scada_file(
    download_url, down_load_to
):
    r = requests.get(download_url, headers=USR_AGENT_HEADER)
    zipped_file = zipfile.ZipFile(io.BytesIO(r.content))

    file_name = zipped_file.namelist()[
        0
    ]  # Just one file so we can pull it out of the list using 0
    start_row_second_table = _find_start_row_nth_table(
        zipped_file, file_name, 1
    )
    csv_file = zipped_file.open(file_name)
    data = pd.read_csv(
        csv_file, header=1, dtype=str
    )[:-1]
    data.to_csv(
        os.path.join(
            down_load_to,
            "PUBLIC_NEXT_DAY_INTERMITTENT_GEN_SCADA_" + file_name[39:47] + ".csv",
        ),
        index=False,
    )


def _find_start_row_nth_table(sub_folder_zipfile, file_name, n):
    row = 0
    table_start_rows_found = 0
    with sub_folder_zipfile.open(file_name) as f:
        for line in f:
            row += 1
            if str(line)[2] == "I":
                table_start_rows_found += 1
                table_start_row = row
                if table_start_rows_found == n:
                    return table_start_row
    raise custom_errors.DataFormatError(
        "The data in table BIDMOVE_COMPLETE was not in the expected format. \n"
        + "Please contact the NEMOSIS package maintainers."
    )



def run_fcas4s(year, month, day, chunk, index, filename_stub, down_load_to):
    """This function"""

    # Add the year and month information to the generic AEMO data url
    url_formatted_latest = defaults.fcas_4_url.format(year, month, day, index)
    url_formatted_hist = defaults.fcas_4_url_hist.format(
        year, year, month, year, month, day, index
    )
    # Perform the download, unzipping saving of the file
    try:
        download_unzip_csv(url_formatted_latest, down_load_to)
    except Exception:
        try:
            download_unzip_csv(url_formatted_hist, down_load_to)
        except Exception as e:
            # FCAS csvs are bundled in 30 minute bundles
            # Check if the csv exists before warning
            file_check = os.path.join(down_load_to, filename_stub + ".csv")
            if not os.path.isfile(file_check):
                logger.warning(f"{filename_stub} not downloaded")


def download_unzip_csv(url, down_load_to):
    """
    This function downloads a zipped csv using a url,
    extracts the csv and saves it a specified location
    """
    url = url.replace('#', '%23')
    r = requests.get(url, headers=USR_AGENT_HEADER)
    z = zipfile.ZipFile(io.BytesIO(r.content))
    z.extractall(down_load_to)


def download_csv(url, path_and_name):
    """
    This function downloads a zipped csv using a url,
    extracts the csv and saves it a specified location
    """
    r = requests.get(url, headers=USR_AGENT_HEADER)
    with open(path_and_name, "wb") as f:
        f.write(r.content)


def download_elements_file(url, path_and_name):
    page = requests.get(url)
    text = page.text
    soup = BeautifulSoup(text, "html.parser")
    links = soup.find_all("a")
    last_file_name = links[-1].text
    link = url + last_file_name
    r = requests.get(link, headers=USR_AGENT_HEADER)
    with open(path_and_name, "wb") as f:
        f.write(r.content)


def download_xl(url, path_and_name):
    """
    This function downloads a zipped csv using a url, extracts the csv and
    saves it a specified location
    """
    r = requests.get(url, headers=USR_AGENT_HEADER)
    with open(path_and_name, "wb") as f:
        f.write(r.content)


def format_aemo_url(url, year, month, filename_stub):
    """
    This fills in the missing information in the AEMO URL
    so data for the right month, year and file name are
    downloaded
    """
    year = str(year)
    return url.format(year, year, month, filename_stub)


def status_code_return(url):
    r = requests.get(url, headers=USR_AGENT_HEADER)
    return r.status_code


def _get_matching_link(url, stub_link):
    r = requests.get(url, headers=USR_AGENT_HEADER)
    soup = BeautifulSoup(r.content, "html.parser")
    links = [link.get("href") for link in soup.find_all("a")]
    for link in links:
        if stub_link in link:
            return link
    logger.warning(f"{stub_link} not downloaded")



import logging
import sys
from .value_parser import _parse_datetime, _parse_column, _infer_column_data_types
from .data_fetch_methods import *

name = "osdan"

logging.getLogger(__name__).addHandler(logging.NullHandler())
logging.basicConfig(
    stream=sys.stdout, level=logging.INFO, format="%(levelname)s: %(message)s"
)



from nemosis import query_wrappers, downloader, date_generators, write_file_names, filters

setup = {
    "DISPATCHLOAD": None,
    "NEXT_DAY_DISPATCHLOAD": None,
    "INTERMITTENT_GEN_SCADA": None,
    "TRADINGLOAD": None,
    "TRADINGPRICE": None,
    "TRADINGREGIONSUM": None,
    "TRADINGINTERCONNECT": None,
    "DISPATCHPRICE": None,
    "DISPATCH_UNIT_SCADA": None,
    "DISPATCHCONSTRAINT": None,
    "DUDETAILSUMMARY": None,
    "PARTICIPANT": None,
    "DUDETAIL": None,
    "GENCONDATA": None,
    "SPDREGIONCONSTRAINT": None,
    "SPDCONNECTIONPOINTCONSTRAINT": None,
    "SPDINTERCONNECTORCONSTRAINT": None,
    "FCAS_4_SECOND": None,
    "ELEMENTS_FCAS_4_SECOND": None,
    "VARIABLES_FCAS_4_SECOND": None,
    "Generators and Scheduled Loads": None,
    "BIDDAYOFFER_D": query_wrappers.dispatch_date_setup,
    "BIDPEROFFER_D": None,
    "FCAS_4s_SCADA_MAP": None,
    "DISPATCHINTERCONNECTORRES": None,
    "DISPATCHREGIONSUM": None,
    "LOSSMODEL": None,
    "LOSSFACTORMODEL": None,
    "MNSP_DAYOFFER": query_wrappers.dispatch_date_setup,
    "MNSP_PEROFFER": query_wrappers.dispatch_half_hour_setup,
    "MNSP_INTERCONNECTOR": None,
    "INTERCONNECTOR": None,
    "INTERCONNECTORCONSTRAINT": None,
    "MARKET_PRICE_THRESHOLDS": None,
    "DAILY_REGION_SUMMARY": None,
    "ROOFTOP_PV_ACTUAL": None,
}

search_type = {
    "DISPATCHLOAD": "start_to_end",
    "NEXT_DAY_DISPATCHLOAD": "start_to_end",
    "INTERMITTENT_GEN_SCADA": "start_to_end",
    "TRADINGLOAD": "start_to_end",
    "TRADINGPRICE": "start_to_end",
    "TRADINGREGIONSUM": "start_to_end",
    "TRADINGINTERCONNECT": "start_to_end",
    "DISPATCHPRICE": "start_to_end",
    "DISPATCH_UNIT_SCADA": "start_to_end",
    "DISPATCHCONSTRAINT": "start_to_end",
    "DUDETAILSUMMARY": "end",
    "PARTICIPANT": "all",
    "DUDETAIL": "all",
    "GENCONDATA": "all",
    "SPDREGIONCONSTRAINT": "all",
    "SPDCONNECTIONPOINTCONSTRAINT": "all",
    "SPDINTERCONNECTORCONSTRAINT": "all",
    "FCAS_4_SECOND": "start_to_end",
    "ELEMENTS_FCAS_4_SECOND": None,
    "VARIABLES_FCAS_4_SECOND": None,
    "Generators and Scheduled Loads": None,
    "BIDDAYOFFER_D": "start_to_end",
    "BIDPEROFFER_D": "start_to_end",
    "FCAS_4s_SCADA_MAP": None,
    "DISPATCHINTERCONNECTORRES": "start_to_end",
    "DISPATCHREGIONSUM": "start_to_end",
    "LOSSMODEL": "all",
    "LOSSFACTORMODEL": "all",
    "MNSP_DAYOFFER": "start_to_end",
    "MNSP_PEROFFER": "start_to_end",
    "MNSP_INTERCONNECTOR": "all",
    "INTERCONNECTOR": "all",
    "INTERCONNECTORCONSTRAINT": "all",
    "MARKET_PRICE_THRESHOLDS": "all",
    "DAILY_REGION_SUMMARY": 'start_to_end',
    "ROOFTOP_PV_ACTUAL": "start_to_end",
}

date_cols = {
    "DISPATCHLOAD": ["SETTLEMENTDATE"],
    "NEXT_DAY_DISPATCHLOAD": ["SETTLEMENTDATE"],
    "INTERMITTENT_GEN_SCADA": ["RUN_DATETIME"],
    "TRADINGLOAD": ["SETTLEMENTDATE"],
    "TRADINGPRICE": ["SETTLEMENTDATE"],
    "TRADINGREGIONSUM": ["SETTLEMENTDATE"],
    "TRADINGINTERCONNECT": ["SETTLEMENTDATE"],
    "DISPATCHPRICE": ["SETTLEMENTDATE"],
    "DISPATCH_UNIT_SCADA": ["SETTLEMENTDATE"],
    "DISPATCHCONSTRAINT": ["SETTLEMENTDATE"],
    "DUDETAILSUMMARY": ["START_DATE", "END_DATE"],
    "PARTICIPANT": ["LASTCHANGED"],
    "DUDETAIL": ["EFFECTIVEDATE"],
    "GENCONDATA": ["EFFECTIVEDATE"],
    "SPDREGIONCONSTRAINT": ["EFFECTIVEDATE"],
    "SPDCONNECTIONPOINTCONSTRAINT": ["EFFECTIVEDATE"],
    "SPDINTERCONNECTORCONSTRAINT": ["EFFECTIVEDATE"],
    "FCAS_4_SECOND": ["TIMESTAMP"],
    "ELEMENTS_FCAS_4_SECOND": None,
    "VARIABLES_FCAS_4_SECOND": None,
    "Generators and Scheduled Loads": None,
    "BIDDAYOFFER_D": ["SETTLEMENTDATE"],
    "BIDPEROFFER_D": ["INTERVAL_DATETIME"],
    "FCAS_4s_SCADA_MAP": None,
    "DISPATCHINTERCONNECTORRES": ["SETTLEMENTDATE"],
    "DISPATCHREGIONSUM": ["SETTLEMENTDATE"],
    "LOSSMODEL": ["EFFECTIVEDATE"],
    "LOSSFACTORMODEL": ["EFFECTIVEDATE"],
    "MNSP_DAYOFFER": ["SETTLEMENTDATE"],
    "MNSP_PEROFFER": ["SETTLEMENTDATE", "PERIODID"],
    "MNSP_INTERCONNECTOR": ["EFFECTIVEDATE"],
    "INTERCONNECTOR": ["LASTCHANGED"],
    "INTERCONNECTORCONSTRAINT": ["EFFECTIVEDATE"],
    "MARKET_PRICE_THRESHOLDS": ["EFFECTIVEDATE"],
    "DAILY_REGION_SUMMARY": ['SETTLEMENTDATE'],
    "ROOFTOP_PV_ACTUAL": ["INTERVAL_DATETIME"],
}

filter = {
    "DISPATCHLOAD": filters.filter_on_settlementdate,
    "NEXT_DAY_DISPATCHLOAD": filters.filter_on_settlementdate,
    "INTERMITTENT_GEN_SCADA": filters.filter_on_run_datetime,
    "TRADINGLOAD": filters.filter_on_settlementdate,
    "TRADINGPRICE": filters.filter_on_settlementdate,
    "TRADINGREGIONSUM": filters.filter_on_settlementdate,
    "TRADINGINTERCONNECT": filters.filter_on_settlementdate,
    "DISPATCHPRICE": filters.filter_on_settlementdate,
    "DISPATCH_UNIT_SCADA": filters.filter_on_settlementdate,
    "DISPATCHCONSTRAINT": filters.filter_on_settlementdate,
    "DUDETAILSUMMARY": filters.filter_on_start_and_end_date,
    "PARTICIPANT": filters.filter_on_last_changed,
    "DUDETAIL": filters.filter_on_effective_date,
    "GENCONDATA": filters.filter_on_effective_date,
    "SPDREGIONCONSTRAINT": filters.filter_on_effective_date,
    "SPDCONNECTIONPOINTCONSTRAINT": filters.filter_on_effective_date,
    "SPDINTERCONNECTORCONSTRAINT": filters.filter_on_effective_date,
    "FCAS_4_SECOND": filters.filter_on_timestamp,
    "ELEMENTS_FCAS_4_SECOND": None,
    "VARIABLES_FCAS_4_SECOND": None,
    "Generators and Scheduled Loads": None,
    "BIDDAYOFFER_D": filters.filter_on_settlementdate,
    "BIDPEROFFER_D": filters.filter_on_interval_datetime,
    "FCAS_4s_SCADA_MAP": None,
    "DISPATCHINTERCONNECTORRES": filters.filter_on_settlementdate,
    "DISPATCHREGIONSUM": filters.filter_on_settlementdate,
    "LOSSMODEL": filters.filter_on_effective_date,
    "LOSSFACTORMODEL": filters.filter_on_effective_date,
    "MNSP_DAYOFFER": filters.filter_on_settlementdate,
    "MNSP_PEROFFER": filters.filter_on_date_and_peroid,
    "MNSP_INTERCONNECTOR": filters.filter_on_effective_date,
    "INTERCONNECTOR": filters.filter_on_last_changed,
    "INTERCONNECTORCONSTRAINT": filters.filter_on_effective_date,
    "MARKET_PRICE_THRESHOLDS": filters.filter_on_effective_date,
    "DAILY_REGION_SUMMARY": filters.filter_on_settlementdate,
    "ROOFTOP_PV_ACTUAL": filters.filter_on_interval_datetime,
}

finalise = {
    "DISPATCHLOAD": None,
    "NEXT_DAY_DISPATCHLOAD": None,
    "INTERMITTENT_GEN_SCADA": None,
    "TRADINGLOAD": None,
    "TRADINGPRICE": None,
    "TRADINGREGIONSUM": None,
    "TRADINGINTERCONNECT": None,
    "DISPATCHPRICE": None,
    "DISPATCH_UNIT_SCADA": None,
    "DISPATCHCONSTRAINT": [
        query_wrappers.convert_genconid_effectivedate_to_datetime_format
    ],
    "DUDETAILSUMMARY": [
        query_wrappers.most_recent_records_before_start_time,
        query_wrappers.drop_duplicates_by_primary_key,
    ],
    "PARTICIPANT": [
        query_wrappers.most_recent_records_before_start_time,
        query_wrappers.drop_duplicates_by_primary_key,
    ],
    "DUDETAIL": [
        query_wrappers.most_recent_records_before_start_time,
        query_wrappers.drop_duplicates_by_primary_key,
    ],
    "GENCONDATA": [
        query_wrappers.most_recent_records_before_start_time,
        query_wrappers.drop_duplicates_by_primary_key,
    ],
    "SPDREGIONCONSTRAINT": [query_wrappers.most_recent_records_before_start_time],
    "SPDCONNECTIONPOINTCONSTRAINT": [
        query_wrappers.most_recent_records_before_start_time,
        query_wrappers.drop_duplicates_by_primary_key,
    ],
    "SPDINTERCONNECTORCONSTRAINT": [
        query_wrappers.most_recent_records_before_start_time,
        query_wrappers.drop_duplicates_by_primary_key,
    ],
    "FCAS_4_SECOND": [query_wrappers.fcas4s_finalise],
    "ELEMENTS_FCAS_4_SECOND": None,
    "VARIABLES_FCAS_4_SECOND": None,
    "Generators and Scheduled Loads": None,
    "BIDDAYOFFER_D": None,
    "BIDPEROFFER_D": None,
    "FCAS_4s_SCADA_MAP": None,
    "DISPATCHINTERCONNECTORRES": None,
    "DISPATCHREGIONSUM": None,
    "LOSSMODEL": [
        query_wrappers.most_recent_records_before_start_time,
        query_wrappers.drop_duplicates_by_primary_key,
    ],
    "LOSSFACTORMODEL": [
        query_wrappers.most_recent_records_before_start_time,
        query_wrappers.drop_duplicates_by_primary_key,
    ],
    "MNSP_DAYOFFER": [
        query_wrappers.most_recent_records_before_start_time,
        query_wrappers.drop_duplicates_by_primary_key,
    ],
    "MNSP_PEROFFER": [
        query_wrappers.most_recent_records_before_start_time,
        query_wrappers.drop_duplicates_by_primary_key,
    ],
    "MNSP_INTERCONNECTOR": [
        query_wrappers.most_recent_records_before_start_time,
        query_wrappers.drop_duplicates_by_primary_key,
    ],
    "INTERCONNECTOR": [
        query_wrappers.most_recent_records_before_start_time,
        query_wrappers.drop_duplicates_by_primary_key,
    ],
    "INTERCONNECTORCONSTRAINT": [
        query_wrappers.most_recent_records_before_start_time,
        query_wrappers.drop_duplicates_by_primary_key,
    ],
    "MARKET_PRICE_THRESHOLDS": None,
    "DAILY_REGION_SUMMARY": None,
    "ROOFTOP_PV_ACTUAL": [
        query_wrappers.drop_duplicates_by_primary_key
    ],
}

date_gen = {
    "MMS": date_generators.year_and_month_gen,
    "NEXT_DAY_DISPATCHLOAD": date_generators.current_gen,
    "INTERMITTENT_GEN_SCADA": date_generators.current_gen,
    "BIDDING": date_generators.bid_table_gen,
    "DAILY_REGION_SUMMARY": date_generators.current_gen,
    "FCAS": date_generators.year_month_day_index_gen,
}

write_filename = {
    "MMS": write_file_names.write_mms_file_names,
    "NEXT_DAY_DISPATCHLOAD": write_file_names.write_file_names_current,
    "INTERMITTENT_GEN_SCADA": write_file_names.write_file_names_current,
    "BIDDING": write_file_names.write_file_names_mms_and_current,
    "DAILY_REGION_SUMMARY": write_file_names.write_file_names_mms_and_current,
    "FCAS": write_file_names.write_file_names_fcas,
}

downloader = {
    "MMS": downloader.run,
    "NEXT_DAY_DISPATCHLOAD": downloader.run_next_dispatch_tables,
    "INTERMITTENT_GEN_SCADA": downloader.run_intermittent_gen_scada,
    "BIDDING": downloader.run_bid_tables,
    "DAILY_REGION_SUMMARY": downloader.run_next_day_region_tables,
    "FCAS": downloader.run_fcas4s,
}
